from .datparser import DATParser
from .dupeparser import DupeParser
from .gamefinder import GameFinder
from .rahasher import RAHasher
from .romchooser import ROMChooser
from .romcleaner import ROMCleaner
from .romdownloader import ROMDownloader
from .rommover import ROMMover
from .romparser import ROMParser
from .rompatcher import ROMPatcher
from .romsearch import ROMSearch

__all__ = [
    "DATParser",
    "DupeParser",
    "GameFinder",
    "RAHasher",
    "ROMChooser",
    "ROMCleaner",
    "ROMDownloader",
    "ROMMover",
    "ROMParser",
    "ROMPatcher",
    "ROMSearch",
]



from .discord import discord_push
from .general import split, match_retool_search_terms, normalize_name, get_file_time
from .io import load_yml, save_yml, unzip_file, load_json, save_json
from .logger import setup_logger, centred_string, left_aligned_string
from .regex_matching import (
    get_file_pattern,
    get_bracketed_file_pattern,
    get_directory_name,
    get_short_name,
    get_region_free_name,
)

__all__ = [
    "setup_logger",
    "centred_string",
    "left_aligned_string",
    "load_yml",
    "save_yml",
    "get_bracketed_file_pattern",
    "get_file_pattern",
    "get_directory_name",
    "get_short_name",
    "get_region_free_name",
    "load_json",
    "save_json",
    "unzip_file",
    "discord_push",
    "split",
    "get_file_time",
    "normalize_name",
    "match_retool_search_terms",
]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "romsearch"
copyright = "2024, bbtufty"
author = "bbtufty"
release = "0.1.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_automodapi.automodapi",
]

templates_path = ["_templates"]
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
]

master_doc = "index"

todo_include_todos = True

html_theme_options = {
    "collapse_navigation": False,
    "navigation_depth": 4,
    "globaltoc_collapse": False,
    "globaltoc_includehidden": False,
    "version_selector": True,
}

autoclass_content = "both"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = []



import os

from ..util import load_yml, load_json


def check_rahasher_names(config_file, platform):
    """Using a config file and platform, get all applicable RA names with hashes

    Args:
        config_file (string): Path to the config file
        platform (string): Platform name
    """

    config = load_yml(config_file)

    ra_hash_dir = config.get("dirs", {}).get("ra_hash_dir", None)

    if ra_hash_dir is None:
        raise ValueError("No ra_hash_dir found in config file")

    ra_hash_filename = os.path.join(ra_hash_dir, f"{platform}.json")

    if not os.path.exists(ra_hash_filename):
        raise ValueError("No RA hash file found")

    ra_hash_dict = load_json(ra_hash_filename)

    ra_hashes = {}
    for r in ra_hash_dict:
        for h in ra_hash_dict[r]["Hashes"]:
            ra_hashes[h["MD5"]] = {
                "name": h["Name"],
                "hash": h["MD5"],
                "patch_name": h["PatchUrl"],
            }

    return ra_hashes



from romsearch import ROMParser

TEST_NAME = "Example Game (USA) (En,De,Fr,Es+It)"


def test_romparser_regions():
    """Put a filename into ROMParser and check it returns the right regions"""

    expected_regions = ["USA"]

    test_case = {TEST_NAME: {"priority": 1}}

    rp = ROMParser(
        config_file="test_config.yml",
        platform="Nintendo - Super Nintendo Entertainment System",
        game="Example Game",
    )
    roms_parsed = rp.run(test_case)

    assert roms_parsed[TEST_NAME]["regions"] == expected_regions


def test_romparser_languages():
    """Put a filename into ROMParser and check it returns the right languages"""

    expected_languages = ["English", "French", "German", "Italian", "Spanish"]

    test_case = {TEST_NAME: {"priority": 1}}

    rp = ROMParser(
        config_file="test_config.yml",
        platform="Nintendo - Super Nintendo Entertainment System",
        game="Example Game",
    )
    roms_parsed = rp.run(test_case)

    assert roms_parsed[TEST_NAME]["languages"] == expected_languages



from .gamefinder_dev import check_regex_parsing, parse_games_from_dat
from .rahasher_dev import check_rahasher_names

__all__ = [
    "check_regex_parsing",
    "check_rahasher_names",
    "parse_games_from_dat",
]



import copy
import glob
import numpy as np
import os
import shutil

import romsearch
from ..util import (
    centred_string,
    load_yml,
    setup_logger,
    load_json,
    save_json,
)


class ROMCleaner:

    def __init__(
        self,
        platform,
        config_file=None,
        config=None,
        platform_config=None,
        logger=None,
        log_line_sep="=",
        log_line_length=100,
    ):
        """ROM directory cleaner

        Cleans out ROMs we no longer want in the final ROM directory

        Args:
            platform (str): Platform name
            config_file (str, optional): path to config file. Defaults to None.
            config (dict, optional): configuration dictionary. Defaults to None.
            platform_config (dict, optional): platform configuration dictionary. Defaults to None.
            logger (logging.Logger, optional): logger. Defaults to None.
            log_line_length (int, optional): Line length of log. Defaults to 100
        """

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get(
                "log_dir", os.path.join(os.getcwd(), "logs")
            )
            logger_add_dir = copy.deepcopy(platform)
            log_level = self.config.get("logger", {}).get("level", "info")
            logger = setup_logger(
                log_level=log_level,
                script_name=f"ROMCleaner",
                log_dir=log_dir,
                additional_dir=logger_add_dir,
            )
        self.logger = logger

        self.rom_dir = self.config.get("dirs", {}).get("rom_dir", None)
        if self.rom_dir is None:
            raise ValueError("rom_dir needs to be defined in config")

        cache_dir = self.config.get("dirs", {}).get("cache_dir", os.getcwd())
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        cache_file = os.path.join(cache_dir, f"cache ({platform}).json")

        if os.path.exists(cache_file):
            cache = load_json(cache_file)
        else:
            cache = {}

        self.platform = platform
        self.cache_file = cache_file
        self.cache = cache

        # Pull in platform config that we need
        mod_dir = os.path.dirname(romsearch.__file__)

        if platform_config is None:
            platform_config_file = os.path.join(
                mod_dir, "configs", "platforms", f"{platform}.yml"
            )
            platform_config = load_yml(platform_config_file)
        self.platform_config = platform_config

        self.unzip = self.platform_config.get("unzip", False)

        self.log_line_sep = log_line_sep
        self.log_line_length = log_line_length

    def run(
        self,
        rom_dict,
    ):
        """Run the ROMCleaner

        Args:
            rom_dict (dict): Dictionary of ROMs we want to keep
        """

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")
        self.logger.info(
            centred_string("Running ROMCleaner", total_length=self.log_line_length)
        )
        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        roms_cleaned, cache_cleaned = self.clean_roms(rom_dict)
        self.save_cache()

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        # Join these up into a dictionary
        cleaned = {
            "ROMs": roms_cleaned,
            "Cache": cache_cleaned,
        }

        return cleaned

    def clean_roms(
        self,
        rom_dict,
    ):
        """Check in directory for any ROMs we don't want any more, and delete them

        Args:
            rom_dict (dict): Dictionary of ROMs we want to keep
        """

        full_rom_dir = os.path.join(self.rom_dir, self.platform)

        # Find files on disk
        roms_on_disk = glob.glob(os.path.join(full_rom_dir, "*", "*.*"))

        # Pull out any included ROMs, stripping out the extension
        roms_in_dict = {}
        for r in rom_dict:

            for f in rom_dict[r]:
                if rom_dict[r][f]["excluded"]:
                    continue
                rom_name = os.path.splitext(f)[0]

                # We need to include if things have been patched. Pull that from the cache
                patched = (
                    self.cache.get(self.platform, {})
                    .get(r, {})
                    .get(f, {})
                    .get("patched", False)
                )
                if patched:
                    rom_name += " (ROMPatched)"

                roms_in_dict[rom_name] = {"full_name": f}

        roms_cleaned = []
        cache_cleaned = []
        dict_cleaned = {}

        for rom_on_disk in roms_on_disk:

            # Pull out a short ROM name, skipping any extensions and parent directories
            rom_base = os.path.split(rom_on_disk)[-1]
            rom_short = os.path.splitext(rom_base)[0]

            # Loop over, see if we can find the ROM
            found_rom_in_dict = False
            for r in roms_in_dict:

                if found_rom_in_dict:
                    continue

                if r == rom_short:
                    found_rom_in_dict = True

            # If we haven't found anything, clear out the cache
            if not found_rom_in_dict:

                # Just loop through the dictionary
                found_entry_in_dict = False
                if self.platform in self.cache:

                    for g in self.cache[self.platform]:

                        if found_entry_in_dict:
                            continue

                        for g_i in self.cache[self.platform][g]:

                            if found_entry_in_dict:
                                continue

                            g_i_short = os.path.splitext(g_i)[0]

                            if "(ROMPatched)" in rom_short:
                                g_i_short += " (ROMPatched)"

                            if g_i_short == rom_short:

                                # Also keep info on the dictionary stuff to clean from the cache
                                if g not in dict_cleaned:
                                    dict_cleaned[g] = []
                                dict_cleaned[g].append(g_i)

                                # Remove from the filesystem
                                found_entry_in_dict = True
                                roms_cleaned.append(rom_base)
                                os.remove(rom_on_disk)
                                self.logger.info(
                                    centred_string(
                                        f"Removed {rom_base} from disk",
                                        total_length=self.log_line_length,
                                    )
                                )

        # Now pass through, removing items we no longer need and keeping track
        # of empty cache items
        cache_items_to_pop = []
        if self.platform in self.cache:
            for d in dict_cleaned:
                for d_i in dict_cleaned[d]:
                    self.cache[self.platform][d].pop(d_i, None)
                    if len(self.cache[self.platform][d]) == 0:
                        cache_items_to_pop.append(d)

        # Go through all the cache entries and if we no longer have them on disk, clear those out
        if self.platform in self.cache:
            for d in self.cache[self.platform]:

                d_is_to_remove = []
                for d_i in self.cache[self.platform][d]:
                    found_d_i_on_disk = False

                    d_i_short = os.path.splitext(d_i)[0]

                    # Account for the fact the file might be patched
                    if self.cache[self.platform][d][d_i]["patched"]:
                        d_i_short += " (ROMPatched)"

                    for r in roms_on_disk:

                        if found_d_i_on_disk:
                            continue

                        r_base = os.path.basename(r)
                        r_short = os.path.splitext(r_base)[0]
                        if r_short == d_i_short:
                            found_d_i_on_disk = True

                    if not found_d_i_on_disk:
                        d_is_to_remove.append(d_i)

                for d_i_to_remove in d_is_to_remove:

                    cache_cleaned.append(d_i_to_remove)
                    self.cache[self.platform][d].pop(d_i_to_remove, None)
                    self.logger.info(
                        centred_string(
                            f"Removed {d_i_to_remove} from cache",
                            total_length=self.log_line_length,
                        )
                    )

                    if len(self.cache[self.platform][d]) == 0:
                        cache_items_to_pop.append(d)

        # Clear out any empty cache items
        if self.platform in self.cache:
            for d in cache_items_to_pop:

                cache_cleaned.append(d)
                self.cache[self.platform].pop(d, None)
                self.logger.info(
                    centred_string(
                        f"Removed {d} from cache", total_length=self.log_line_length
                    )
                )

        # Remove any empty directories
        all_dirs = os.listdir(full_rom_dir)
        for d in all_dirs:
            full_dir = os.path.join(full_rom_dir, d)
            if not os.listdir(full_dir):
                shutil.rmtree(full_dir)

        # Because there can be some duplicates, remove here
        roms_cleaned = np.unique(roms_cleaned)
        roms_cleaned = [str(f) for f in roms_cleaned]

        cache_cleaned = np.unique(cache_cleaned)
        cache_cleaned = [str(f) for f in cache_cleaned]

        return roms_cleaned, cache_cleaned

    def save_cache(self):
        """Save out the cache file"""

        cache = copy.deepcopy(self.cache)
        save_json(cache, self.cache_file, sort_key=self.platform)



import copy
import numpy as np
import os
import re
from iso639 import Lang

import romsearch
from .romparser import ROMParser
from ..util import (
    centred_string,
    left_aligned_string,
    setup_logger,
    load_yml,
    load_json,
    match_retool_search_terms,
    get_directory_name,
)

DUPE_DEFAULT = {
    "is_compilation": False,
    "name_type": None,
    "priority": 1,
    "title_pos": None,
    "filters": None,
}


def check_regions(
    filter_regions,
    rom_regions,
):
    """Check condition regions against parsed regions

    Here, we only check that any of the ROM regions match any of the
    filter regions

    Args:
        filter_regions (list): List of regions to match from the filter
        rom_regions (list): Parsed ROM regions
    """

    s_f = set(filter_regions)
    s_r = set(rom_regions)

    s_i = s_r.intersection(s_f)

    # If we've got no matches, then this hasn't been satisfied
    if len(s_i) == 0:
        return False

    return True


def check_languages(
    filter_langs,
    rom_langs,
):
    """Check condition languages against parsed languages

    Here, we only check that any of the ROM languages match any of the
    filter languages. N.B. retool here uses ISO 639-1, so we need to
    parse them

    Args:
        filter_langs (list): List of languages to match from the filter
        rom_langs (list): Parsed ROM languages
    """

    long_filter_langs = [Lang(fl.lower()).name for fl in filter_langs]

    s_f = set(long_filter_langs)
    s_r = set(rom_langs)

    s_i = s_r.intersection(s_f)

    # If we've got no matches, then this hasn't been satisfied
    if len(s_i) == 0:
        return False

    return True


def check_string(
    regex_str,
    file_name,
):
    """Check filename for regex string

    Args:
        regex_str (str): String to check
        file_name (str): Name of file to check
    """

    match = re.search(regex_str, file_name)

    if match is None:
        return False

    return True


def check_region_order(
    region_order,
    user_region_preferences,
    all_regions=None,
):
    """Check region order against parsed regions

    Here, we only check that *any* of the higher regions are
    higher in user preference than *all* of the lower regions

    Args:
        region_order (dict): Dictionary of form {"higherRegions": [], "lowerRegions": []}
        user_region_preferences (list): Ordered list of user region preferences
        all_regions (list): List of all available regions
    """

    if all_regions is None:
        all_regions = []

    higher_regions = region_order["higherRegions"]
    lower_regions = region_order["lowerRegions"]

    # If we've got an 'all other regions' here, then
    # pull them out
    updated_higher_regions = copy.deepcopy(higher_regions)
    if higher_regions == ["All other regions"]:
        updated_higher_regions = copy.deepcopy(all_regions)
        for reg in lower_regions:
            updated_higher_regions.remove(reg)

    updated_lower_regions = copy.deepcopy(lower_regions)
    if lower_regions == ["All other regions"]:
        updated_lower_regions = copy.deepcopy(all_regions)
        for reg in higher_regions:
            updated_lower_regions.remove(reg)

    higher_regions = copy.deepcopy(updated_higher_regions)
    lower_regions = copy.deepcopy(updated_lower_regions)

    high_prio_region = 99
    for reg in higher_regions:

        # If we have the region in the preferences, note the location
        if reg in user_region_preferences:

            # Take the highest priority
            high_prio_region = min(
                [high_prio_region, user_region_preferences.index(reg)]
            )

    # If we still have a high priority of 99, we haven't found anything so jump out
    if high_prio_region == 99:
        return False

    # Now look through the low priorities, and find the highest priority of those
    low_prio_region = 99
    for reg in lower_regions:

        # If we have the region in the preferences, note the location
        if reg in user_region_preferences:

            # Take the highest priority
            low_prio_region = min([low_prio_region, user_region_preferences.index(reg)])

    # If we're at a low priority of 99, then we haven't found anything so we're good to
    # return a True
    if low_prio_region == 99:
        return True

    # If *any* of the high priority regions are higher than *all* of the low
    # priority, then return True. Else, False
    if low_prio_region > high_prio_region:
        return False

    return True


def apply_results(
    parent_name,
    game_dict,
    results,
):
    """Apply results to a filtered match

    Args:
        parent_name: Parent name
        game_dict: Dictionary of game properties
        results: Dictionary of results to apply to the game/game dict
    """

    for r in results:

        # If we're changing the name, that gets pulled out here
        if r == "group":
            parent_name = copy.deepcopy(results[r])

        # If we're changing priority, edit the game dict
        elif r == "priority":
            game_dict[r] = results[r]

        # Ignore local names, since we won't use them
        elif r == "localNames":
            continue

        else:
            raise ValueError(f"Unsure how to deal with result type {r}")

    return parent_name, game_dict


def add_dupe_entry_to_game_dict(
    game,
    game_dict,
    parent,
):
    """Add a dupe entry to the game dict

    Args:
        game (dict): Dictionary containing various game names
        game_dict (dict): Dictionary of games. Defaults
            to None, which will create an empty dict
        parent: Parent for the game. Can either
            be a string, which will set up a default
            dictionary for the parent, or a dictionary
            to inherit
    """

    game_name = copy.deepcopy(game["original_name"])

    if isinstance(parent, str):
        parent_name = copy.deepcopy(parent)
        dupe_entry = DUPE_DEFAULT
    elif isinstance(parent, dict):
        parent_name = parent["parent_name"]
        dupe_entry = parent["dupe_entry"]
    else:
        raise ValueError(f"parent should be one of str, dict")

    # Update the directory name, so that it matches the short parent name
    new_dir_name = get_directory_name(parent_name)
    game["dir_name"] = copy.deepcopy(new_dir_name)

    parent_name_lower = parent_name.lower()
    game_dict_keys = [key for key in game_dict.keys()]
    game_dict_keys_lower = [key.lower() for key in game_dict.keys()]

    if parent_name_lower not in game_dict_keys_lower:
        game_dict[parent_name] = {}
        final_parent_name = copy.deepcopy(parent_name)
    else:
        final_parent_idx = game_dict_keys_lower.index(parent_name_lower)
        final_parent_name = game_dict_keys[final_parent_idx]

    # We want to make sure we also don't duplicate on the names being upper/lowercase
    g_names = [g_dict for g_dict in game_dict[final_parent_name]]
    g_names_lower = [g_name.lower() for g_name in g_names]
    if game_name.lower() in g_names_lower:
        g_idx = g_names_lower.index(game_name.lower())
        game_name = g_names[g_idx]

    if game_name not in game_dict[final_parent_name]:
        game_dict[final_parent_name][game_name] = {}

    # Update with names, then with any dupe entry
    game_dict[final_parent_name][game_name].update(game)
    game_dict[final_parent_name][game_name].update(dupe_entry)

    return game_dict


class GameFinder:

    def __init__(
        self,
        platform,
        dat=None,
        retool=None,
        ra_hashes=None,
        config_file=None,
        config=None,
        dupe_dict=None,
        platform_config=None,
        default_config=None,
        regex_config=None,
        logger=None,
        log_line_sep="=",
        log_line_length=100,
    ):
        """Tool to find games within a list of files

        Will parse through files to get a unique list of games, then pull
        out potential aliases and optionally remove things from user excluded list

        Args:
            platform (str): Platform name
            config_file (str, optional): Path to config file. Defaults to None.
            config (dict, optional): Configuration dictionary. Defaults to None.
            dupe_dict (dict, optional): Dupe dictionary. Defaults to None.
            default_config (dict, optional): Default configuration dictionary. Defaults to None.
            regex_config (dict, optional): Dictionary of regex config. Defaults to None.
            logger (logging.Logger, optional): Logger instance. Defaults to None.
            log_line_length (int, optional): Line length of log. Defaults to 100
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get(
                "log_dir", os.path.join(os.getcwd(), "logs")
            )
            log_level = self.config.get("logger", {}).get("level", "info")
            logger = setup_logger(
                log_level=log_level,
                script_name=f"GameFinder",
                log_dir=log_dir,
                additional_dir=platform,
            )
        self.logger = logger

        # Pull in specifics to include/exclude
        self.include_games = self.config.get("include_games", None)
        self.exclude_games = self.config.get("exclude_games", None)

        # Pull in defaults for finding short game names
        mod_dir = os.path.dirname(romsearch.__file__)

        if default_config is None:
            default_file = os.path.join(mod_dir, "configs", "defaults.yml")
            default_config = load_yml(default_file)
        self.default_config = default_config

        if regex_config is None:
            regex_file = os.path.join(mod_dir, "configs", "regex.yml")
            regex_config = load_yml(regex_file)
        self.regex_config = regex_config

        self.dat = dat
        self.retool = retool
        self.ra_hashes = ra_hashes
        self.platform_config = platform_config

        # Info for dupes
        self.dupe_dict = dupe_dict
        self.dupe_dir = config.get("dirs", {}).get("dupe_dir", None)
        self.filter_dupes = config.get("gamefinder", {}).get("filter_dupes", True)

        self.log_line_sep = log_line_sep
        self.log_line_length = log_line_length

    def run(
        self,
        files,
    ):

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")
        self.logger.info(
            centred_string("Running GameFinder", total_length=self.log_line_length)
        )
        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        games_dict = self.get_game_dict(files)
        games_dict = dict(sorted(games_dict.items()))

        self.logger.info(
            centred_string(
                f"Found {len(games_dict)} games", total_length=self.log_line_length
            )
        )
        self.logger.debug(f"{'-' * self.log_line_length}")
        for gi, g in enumerate(games_dict):
            self.logger.debug(
                left_aligned_string(f"{g}:", total_length=self.log_line_length)
            )

            for ga in games_dict[g]:
                self.logger.debug(
                    left_aligned_string(
                        f"-> Priority {games_dict[g][ga]['priority']}: {ga}",
                        total_length=self.log_line_length,
                    )
                )

            if gi != len(games_dict) - 1:
                self.logger.debug(f"{'-' * self.log_line_length}")

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        return games_dict

    def get_game_dict(
        self,
        files,
    ):
        """Get a game dictionary out.

        From a list of files, parse out dupes, apply includes and excludes,
        and return a game dictionary.

        Args:
            files (dict): Dictionary of files with names for association
        """

        # We need to trim down dupes here. Otherwise, the
        # dict is just the list we already have
        game_dict = None
        if self.filter_dupes:
            game_dict = self.get_filter_dupes(files)

        # If the dupe filtering has failed, then just assign all the short names to the default dupe dict
        if game_dict is None:

            game_dict = {}

            for f in files:
                short_name = copy.deepcopy(files[f]["short_name"])
                game_dict[short_name] = {
                    short_name: DUPE_DEFAULT,
                }

        # Remove any excluded files
        if self.exclude_games is not None:
            games_to_remove = self.get_game_matches(
                game_dict,
                self.exclude_games,
            )
            if games_to_remove is not None:
                for g in games_to_remove:
                    del game_dict[g]

        # Include only included files
        if self.include_games is not None:

            games_to_include = self.get_game_matches(
                game_dict,
                self.include_games,
            )
            if games_to_include is not None:

                filtered_game_dict = {}
                for g in games_to_include:
                    filtered_game_dict[g] = game_dict[g]

                game_dict = copy.deepcopy(filtered_game_dict)
        return game_dict

    def get_game_matches(
        self,
        game_dict,
        games_to_match,
    ):
        """Get files that match an input dictionary (to properly handle dupes)

        Args:
            - game_dict (dict): Dictionary of games to match against
            - games_to_match (list): List of values to match against
        """
        games_matched = []

        if isinstance(games_to_match, dict):
            games_to_match = games_to_match.get(self.platform, None)
        else:
            games_to_match = copy.deepcopy(games_to_match)

        if games_to_match is None:
            return None

        games_matched.extend(games_to_match)

        game_dict_keys = []
        for g in game_dict:
            found_f = False

            for game_matched in games_matched:

                if found_f:
                    continue

                # Look in the group name
                re_find = re.findall(f"^({re.escape(game_matched)}).*", g)

                if len(re_find) > 0:
                    game_dict_keys.append(g)
                    found_f = True

                # If not found, look in the dupe names
                if not found_f:
                    for g_d in game_dict[g]:

                        if found_f:
                            continue

                        re_find = re.findall(f"^({re.escape(game_matched)}).*", g_d)

                        if len(re_find) > 0:
                            game_dict_keys.append(g)
                            found_f = True

        game_dict_keys = np.unique(game_dict_keys)
        game_dict_keys = [str(g) for g in game_dict_keys]

        if len(game_dict_keys) == 0:
            game_dict_keys = None

        return game_dict_keys

    def get_filter_dupes(self, games):
        """Parse down a list of files based on an input dupe list"""

        if self.dupe_dict is None and self.dupe_dir is None:
            raise ValueError(
                "dupe_dict or dupe_dir must be specified if filtering dupes"
            )

        if self.dupe_dict is None:
            dupe_file = os.path.join(self.dupe_dir, f"{self.platform} (dupes).json")
            if not os.path.exists(dupe_file):
                self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
                self.logger.warning(
                    centred_string(
                        "No dupe files found", total_length=self.log_line_length
                    )
                )
                self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
                return None
            self.dupe_dict = load_json(dupe_file)

        game_dict = {}

        # Loop over games, and the dupes dictionary. Also pull out various other important info
        for i, g in enumerate(games):

            # Look at the search terms
            game_dict = self.filter_by_search_terms(
                game=games[g],
                game_dict=game_dict,
            )

        return game_dict

    def filter_by_search_terms(
        self,
        game,
        game_dict=None,
    ):
        """Add entries to game dict based on search term conditions

        Will find possible parents, then add those to the game dict

        Args:
            game (dict): Dictionary containing various game names
            game_dict (dict): Dictionary of games to match against. Defaults
                to None, which will create an empty dict
        """

        if game_dict is None:
            game_dict = {}

        found_parents = self.get_parents(
            game_name=game,
        )

        for parent in found_parents:

            # If we have multiple matches, and it's not part of a compilation, freak out
            if len(found_parents) > 1:
                is_compilation = parent.get("dupe_entry").get("is_compilation", False)
                if not is_compilation:
                    self.logger.warning(
                        centred_string(
                            f"{game['original_name']} has multiple matches! "
                            f"This should not generally happen",
                            total_length=self.log_line_length,
                        )
                    )

            game_dict = add_dupe_entry_to_game_dict(
                game=game,
                game_dict=game_dict,
                parent=parent,
            )

        return game_dict

    def get_parents(
        self,
        game_name,
    ):
        """Get the parent(s) recursively searching through a dupe dict

        Because we can have compilations, find all cases where things match up

        Args:
            game_name (dict): game name to find parents for
        """

        # Keep track of the fact that we might have region-free names here
        if isinstance(game_name, dict):
            full_name = copy.deepcopy(game_name["full_name"])
            short_name = copy.deepcopy(game_name["short_name"])
            region_free_name = copy.deepcopy(game_name["region_free_name"])
        else:
            full_name = copy.deepcopy(game_name)
            short_name = copy.deepcopy(game_name)
            region_free_name = copy.deepcopy(game_name)

        # We do this by lowercase checking
        dupes = [{g: self.dupe_dict[g]} for g in self.dupe_dict]

        # Pull out all the clones (and whether they've got name types) so we can check that way as well
        all_clones = []
        name_types = []

        for d in self.dupe_dict:
            all_clones.append(
                {key: self.dupe_dict[d][key] for key in self.dupe_dict[d]}
            )

        for clone in all_clones:
            name_type = [clone[k].get("name_type", None) for k in clone]
            name_types.append(name_type)

        found_dupe = False

        found_parents = []

        # Check all the clones within the dupes. These can potentially
        # have different name types, so loop
        for clone, name_type, dupe in zip(all_clones, name_types, dupes):

            for c, n in zip(clone, name_type):

                found_dupe_in_clone = match_retool_search_terms(
                    full_name=full_name,
                    search_term=c,
                    short_name=short_name,
                    region_free_name=region_free_name,
                    match_type=n,
                )
                if found_dupe_in_clone:
                    found_dupe = True

                if found_dupe_in_clone:

                    parent_name = list(dupe.keys())[0]
                    dupe_entry = copy.deepcopy(clone[c])

                    # We need to apply filters here, hey
                    filters = clone[c].get("filters", None)
                    if filters is not None:

                        parent_name, dupe_entry = self.apply_filters(
                            parent_name=parent_name,
                            dupe_entry=dupe_entry,
                            game_name=game_name,
                            filters=filters,
                        )

                    # Set up the dictionary
                    found_parent = {
                        "parent_name": parent_name,
                        "dupe_name": c,
                        "dupe_entry": dupe_entry,
                    }

                    # Don't duplicate
                    if found_parent not in found_parents:
                        found_parents.append(found_parent)

        if not found_dupe:
            found_parents = copy.deepcopy(short_name)

        if found_parents is None:
            raise ValueError("Could not find a parent name!")

        if not isinstance(found_parents, list):
            found_parents = [found_parents]

        return found_parents

    def apply_filters(
        self,
        game_name,
        parent_name,
        dupe_entry,
        filters,
    ):
        """Apply filters to game if conditions are met

        Args:
            game_name: Dictionary containing various game names
            parent_name: Game name
            dupe_entry: Dictionary of game properties
            filters: Filters with conditions and results to apply
        """

        # Parse out the filename
        file_to_parse = {game_name["full_name"]: game_name}
        rp = ROMParser(
            platform=self.platform,
            game=parent_name,
            config=self.config,
            dat=self.dat,
            retool=self.retool,
            ra_hashes=self.ra_hashes,
            default_config=self.default_config,
            regex_config=self.regex_config,
            logger=self.logger,
        )
        rom_parsed = rp.run(file_to_parse)
        rom_parsed = rom_parsed.get(game_name["full_name"], None)

        if rom_parsed is None:
            raise ValueError("ROM parsing failed")

        # Having parsed the file, now loop over the conditions
        for filt in filters:

            conditions_met = []
            for c in filt["conditions"]:
                if c == "matchLanguages":
                    condition_met = check_languages(
                        filt["conditions"][c], rom_parsed["languages"]
                    )
                elif c == "matchRegions":
                    condition_met = check_regions(
                        filt["conditions"][c], rom_parsed["regions"]
                    )
                elif c == "matchString":
                    condition_met = check_string(
                        filt["conditions"][c], game_name["full_name"]
                    )
                elif c == "regionOrder":
                    condition_met = check_region_order(
                        filt["conditions"][c],
                        self.config["region_preferences"],
                        all_regions=list(self.default_config["regions"]),
                    )
                else:
                    raise ValueError(f"Unsure how to deal with condition {c}")
                conditions_met.append(condition_met)

            # If we've met the conditions, then apply the results
            if all(conditions_met):
                parent_name, dupe_entry = apply_results(
                    parent_name,
                    dupe_entry,
                    filt["results"],
                )

        return parent_name, dupe_entry



import copy
import glob
import os
import re
import subprocess

import romsearch
from ..util import (
    setup_logger,
    load_yml,
    get_file_pattern,
    discord_push,
    split,
    centred_string,
)

RCLONE_METHODS = [
    "sync",
    "copy",
]


def add_rclone_filter(
    pattern=None,
    filter_type="include",
    include_wildcard=True,
):
    if filter_type == "include":
        filter_str = "+"
    elif filter_type == "exclude":
        filter_str = "-"
    else:
        raise ValueError("filter_type should be one of include or exclude")

    # rclone wants double curly braces which we need to escape in python strings (yum)
    filter_pattern = ""

    if pattern is not None:
        filter_pattern += f"{{{{{pattern}}}}}"

    if include_wildcard:
        filter_pattern += "*"

    cmd = f'--filter "{filter_str} {filter_pattern}" '

    return cmd


def get_tidy_files(glob_pattern):
    """Get a tidy list of files from a glob pattern.

    This just strips off the leading directories to just get a filename

    Args:
        glob_pattern (str): glob pattern to match
    """

    files = glob.glob(glob_pattern)
    files = [os.path.split(f)[-1] for f in files]

    return files


class ROMDownloader:

    def __init__(
        self,
        platform=None,
        config_file=None,
        config=None,
        platform_config=None,
        rclone_method="sync",
        copy_files=None,
        logger=None,
        include_filter_wildcard=True,
        log_line_sep="=",
        log_line_length=100,
    ):
        """Downloader tool via rclone

        This works per-platform, so must be specified here.

        For this, we can either sync an entire directory (rclone_method='sync'), or copy individual files
        (rclone_method='copy'). If the method is copy, then copy_files must be set as a list

        Args:
            platform (str, optional): Platform name. Defaults to None, which will throw a ValueError
            config_file (str, optional): Configuration file. Defaults to None
            config (dict, optional): Configuration dictionary. Defaults to None
            platform_config (dict, optional): Platform configuration dictionary. Defaults to None
            rclone_method (str, optional): Should be one of 'sync' or 'copy'. Defaults to 'sync'
            copy_files (list, optional): Must be set if rclone_method is 'copy'. Determines the filenames to copy over.
                Defaults to None
            logger (logging.Logger, optional): Logger instance. Defaults to None
            include_filter_wildcard (bool, optional): If set, will include wildcards in rclone filters. Defaults to
                True.
            log_line_length (int, optional): Line length of log. Defaults to 100
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get(
                "log_dir", os.path.join(os.getcwd(), "logs")
            )
            log_level = self.config.get("logger", {}).get("level", "info")
            logger = setup_logger(
                log_level=log_level,
                script_name=f"ROMDownloader",
                log_dir=log_dir,
                additional_dir=platform,
            )
        self.logger = logger

        if rclone_method not in RCLONE_METHODS:
            raise ValueError(f"rclone_method must be one of {RCLONE_METHODS}")

        self.rclone_method = rclone_method
        self.copy_files = copy_files

        out_dir = self.config.get("dirs", {}).get("raw_dir", None)
        if out_dir is None:
            raise ValueError("raw_dir needs to be defined in config")
        self.out_dir = os.path.join(out_dir, platform)

        # Get any specific includes/excludes
        include_games = self.config.get("include_games", None)
        if isinstance(include_games, dict):
            include_games = include_games.get(platform, None)
        else:
            include_games = copy.deepcopy(include_games)

        self.include_games = include_games

        exclude_games = self.config.get("exclude_games", None)
        if isinstance(exclude_games, dict):
            exclude_games = exclude_games.get(platform, None)
        else:
            exclude_games = copy.deepcopy(exclude_games)

        self.exclude_games = exclude_games

        remote_name = self.config.get("romdownloader", {}).get("remote_name", None)
        if remote_name is None:
            raise ValueError("remote_name must be specified in config")
        self.remote_name = remote_name

        sync_all = self.config.get("romdownloader", {}).get("sync_all", True)

        # If we have includes or excludes, force sync all False
        if self.include_games is not None or self.exclude_games is not None:
            sync_all = False

        self.sync_all = sync_all

        self.include_filter_wildcard = include_filter_wildcard

        # Read in the specific platform configuration
        mod_dir = os.path.dirname(romsearch.__file__)

        if platform_config is None:
            platform_config_file = os.path.join(
                mod_dir, "configs", "platforms", f"{platform}.yml"
            )
            platform_config = load_yml(platform_config_file)
        self.platform_config = platform_config

        remote_dir = self.platform_config.get("dir", None)
        if remote_dir is None:
            raise ValueError(f"dir should be defined in the platform config file!")

        # If we're not using absolute URLs, then remove that here
        self.use_absolute_url = self.config.get("romdownloader", {}).get(
            "use_absolute_url", True
        )
        if not self.use_absolute_url:
            if remote_dir[0] == "/":
                remote_dir = remote_dir[1:]

        self.remote_dir = remote_dir

        self.discord_url = self.config.get("discord", {}).get("webhook_url", None)
        self.dry_run = self.config.get("romdownloader", {}).get("dry_run", False)

        self.log_line_sep = log_line_sep
        self.log_line_length = log_line_length

    def run(
        self,
    ):
        """Run Rclone downloader tool"""

        start_files = get_tidy_files(os.path.join(str(self.out_dir), "*"))

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")
        self.logger.info(
            centred_string("Running ROMDownloader", total_length=self.log_line_length)
        )
        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        self.rclone_download(
            remote_dir=self.remote_dir,
            out_dir=self.out_dir,
        )

        end_files = get_tidy_files(os.path.join(str(self.out_dir), "*"))

        if self.discord_url is not None:
            name = f"ROMDownloader: {self.platform}"
            self.post_to_discord(start_files, end_files, name=name)

        # If there are potential additional files to download, do that here
        if "additional_dirs" in self.platform_config:

            for add_dir in self.platform_config["additional_dirs"]:
                add_remote_dir = self.platform_config["additional_dirs"][add_dir]

                # If using relative URL, strip the leading slash
                if not self.use_absolute_url:
                    if add_remote_dir[0] == "/":
                        add_remote_dir = add_remote_dir[1:]

                add_out_dir = f"{self.out_dir} {add_dir}"

                start_files = get_tidy_files(os.path.join(str(add_out_dir), "*"))

                self.rclone_download(
                    remote_dir=add_remote_dir,
                    out_dir=add_out_dir,
                )

                end_files = get_tidy_files(os.path.join(str(add_out_dir), "*"))

                if self.discord_url is not None:
                    name = f"ROMDownloader: {self.platform} ({add_dir})"
                    self.post_to_discord(start_files, end_files, name=name)

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        return True

    def rclone_download(
        self,
        remote_dir,
        out_dir=None,
        max_retries=5,
    ):
        """Download from rclone, either via sync or copy

        Args:
            remote_dir: rclone remote path
            out_dir: directory to download to
            max_retries: maximum number of retries
        """

        if out_dir is None:
            out_dir = os.getcwd()

        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        if self.rclone_method == "sync":
            self.rclone_sync(
                remote_dir=remote_dir,
                out_dir=out_dir,
                max_retries=max_retries,
            )
        elif self.rclone_method == "copy":
            self.rclone_copy(
                remote_dir=remote_dir,
                out_dir=out_dir,
                max_retries=max_retries,
            )
        else:
            raise ValueError(f"rclone_method should be one of {RCLONE_METHODS}")

    def rclone_sync(
        self,
        remote_dir,
        out_dir=None,
        transfers=5,
        max_retries=5,
    ):
        """Use rclone to sync an entire directory

        Args:
            remote_dir: rclone remote path
            out_dir: directory to download to
            transfers: number of simultaneous transfers
            max_retries: maximum number of retries
        """

        # Build the sync command. We take the following steps to avoid errors
        # - Disable HTTP2 to avoid GOAWAY errors
        # - Disable multi-thread transfers

        cmd = (
            f"rclone sync "
            f"--inplace "
            f"--fast-list "
            f"--delete-after "
            f"--disable-http2 "
            f"--multi-thread-streams=0 "
            f"--size-only "
            f"--transfers={transfers} "
            f'"{self.remote_name}:{remote_dir}" '
            f'"{out_dir}" '
            f"-v "
        )

        # Include any filters if necessary (which is probably the case)
        if not self.sync_all:

            # Start with any negative filters
            searches = []

            if self.exclude_games is not None:
                searches.extend(self.exclude_games)

            if len(searches) > 0:
                pattern = get_file_pattern(searches)
            else:
                pattern = None

            if pattern:
                cmd += add_rclone_filter(
                    pattern=pattern,
                    filter_type="exclude",
                    include_wildcard=self.include_filter_wildcard,
                )

            # Now onto positive filters
            searches = []

            # Specific games
            if self.include_games is not None:
                searches.extend(self.include_games)

            if len(searches) > 0:
                pattern = get_file_pattern(searches)
            else:
                pattern = None

            if pattern:
                cmd += add_rclone_filter(
                    pattern=pattern,
                    filter_type="include",
                    include_wildcard=self.include_filter_wildcard,
                )

                cmd += '--filter "- *" '

        if self.dry_run:
            self.logger.info(
                centred_string(
                    f"Dry run, would rclone sync with:",
                    total_length=self.log_line_length,
                )
            )
            self.logger.info(centred_string(cmd, total_length=self.log_line_length))
        else:

            retry = 0
            retcode = 1

            self.logger.info(
                centred_string("Running rclone sync", total_length=self.log_line_length)
            )

            while retcode != 0 and retry < max_retries:

                # Execute the command and capture the output
                with subprocess.Popen(
                    cmd, text=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                ) as process:
                    for line in process.stdout:

                        # Replace any potential tabs in the line, strip whitespace and skip newline at the end
                        line = re.sub("\s+", " ", line[:-1])
                        line = line.lstrip().rstrip()

                        if len(line) == 0:
                            continue

                        # Skip the warning about directory filters using regex
                        if "Can't figure out directory filters" in line:
                            continue

                        # Log each line of the output using the provided logger
                        self.logger.info(
                            centred_string(line, total_length=self.log_line_length)
                        )

                retcode = process.poll()
                retry += 1

            # If we've hit the maximum retries and still we have errors, raise an error with the args
            if retcode != 0:
                raise subprocess.CalledProcessError(retcode, process.args)

        return True

    def rclone_copy(
        self,
        remote_dir,
        out_dir=None,
        max_retries=5,
    ):
        """Use rclone to copy files one-by-one

        Args:
            remote_dir: rclone remote path
            out_dir: directory to download to
            max_retries: maximum number of retries
        """

        if self.copy_files is None:
            raise ValueError("copy_files needs to be defined for rclone copy")

        for fi, f in enumerate(self.copy_files):

            remote_file_name = f"{self.remote_name}:{remote_dir}{f}"

            cmd = (
                f"rclone copy "
                f"--inplace "
                f"--no-traverse "
                f"--disable-http2 "
                f"--multi-thread-streams=0 "
                f"--size-only "
                f'"{remote_file_name}" "{out_dir}" '
                f"-v "
            )

            short_out_dir = os.path.split(out_dir)[-1]

            if self.dry_run:
                self.logger.info(
                    centred_string(
                        f"Dry run, would rclone copy {short_out_dir}, {f} with:",
                        total_length=self.log_line_length,
                    )
                )
                self.logger.info(centred_string(cmd, total_length=self.log_line_length))
            else:

                retry = 0
                retcode = 1

                self.logger.info(
                    centred_string(
                        f"Running rclone copy for {short_out_dir}, {f}",
                        total_length=self.log_line_length,
                    )
                )

                # The retcode is set to 9999 if the file doesn't exist
                while retcode not in [0, 9999] and retry < max_retries:

                    # Execute the command and capture the output
                    with subprocess.Popen(
                        cmd, text=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                    ) as process:
                        for line in process.stdout:

                            # Replace any potential tabs in the line, strip whitespace and skip newline at the end
                            line = re.sub("\s+", " ", line[:-1])
                            line = line.lstrip().rstrip()

                            if len(line) == 0:
                                continue

                            # Skip weird time notifications
                            if "Time may be set wrong" in line:
                                continue

                            # If the file doesn't exist, set a high number and immediately terminate the process
                            if (
                                "directory not found" in line
                                or "object not found" in line
                            ):
                                retcode = 9999
                                process.kill()
                                continue

                            # Log each line of the output using the provided logger
                            self.logger.info(
                                centred_string(
                                    line,  # Exclude the newline character
                                    total_length=self.log_line_length,
                                )
                            )

                    if retcode != 9999:
                        retcode = process.poll()
                    retry += 1

                # If we've hit the maximum retries and still we have errors, raise an error with the args
                if retcode not in [0, 9999]:
                    raise subprocess.CalledProcessError(retcode, process.args)
                if retcode == 9999:
                    self.logger.warning(
                        centred_string(
                            f"Could not find {self.remote_name}:{remote_dir}{f}. "
                            f"This may not be an issue",
                            total_length=self.log_line_length,
                        )
                    )

            if fi != len(self.copy_files) - 1:
                self.logger.info(f"{'-' * self.log_line_length}")

        # Do a pass through where we delete all extraneous files at the end
        # If we're checking files, then do a pass where if we don't find the file in the includes then
        # we delete it
        all_files = get_tidy_files(os.path.join(str(out_dir), "*"))

        found_matches = []

        for f in all_files:

            found_match = False

            for c in self.copy_files:
                if f == c:
                    found_match = True

            if not found_match:
                os.remove(os.path.join(str(out_dir), f))
                found_matches.append(f)

        # Do a nice summary of what we might have removed
        if len(found_matches) > 0:
            self.logger.info(f"{'-' * self.log_line_length}")
            self.logger.info(
                centred_string(f"Removed files:", total_length=self.log_line_length)
            )
            for f in found_matches:
                self.logger.info(
                    centred_string(f"{f}", total_length=self.log_line_length)
                )

        return True

    def post_to_discord(
        self,
        start_files,
        end_files,
        name,
        max_per_message=10,
    ):
        """Create a discord post summarising files added and removed

        Args:
            start_files (list): list of files at the start of the rclone
            end_files (list): list of files at the end of the rclone
            name (string): Name of the post title
            max_per_message (int, optional): Maximum number of items per post. Defaults to 10.
        """

        items_added = list(set(end_files).difference(start_files))
        items_deleted = list(set(start_files).difference(end_files))

        if len(items_added) > 0:

            items_added.sort()

            for items_split in split(items_added, chunk_size=max_per_message):

                fields = []

                field_dict = {"name": "Added", "value": "\n".join(items_split)}
                fields.append(field_dict)

                if len(fields) > 0:
                    discord_push(
                        url=self.discord_url,
                        name=name,
                        fields=fields,
                    )

        if len(items_deleted) > 0:

            items_deleted.sort()

            for items_split in split(items_deleted, chunk_size=max_per_message):

                fields = []

                field_dict = {"name": "Deleted", "value": "\n".join(items_split)}
                fields.append(field_dict)

                if len(fields) > 0:
                    discord_push(
                        url=self.discord_url,
                        name=name,
                        fields=fields,
                    )

        return True



import logging
import os

import romsearch
from romsearch import GameFinder
from ..util import (
    get_short_name,
    load_json,
    load_yml,
    get_directory_name,
    normalize_name,
    get_region_free_name,
)


def check_regex_parsing(dat_filename):
    """Run a .dat file through the short name parser

    This is useful when adding a new platform, so we can see
    what new terms need to be added
    """

    dat_file = load_json(dat_filename)
    all_files = [f for f in dat_file]
    all_files.sort()

    mod_dir = os.path.dirname(romsearch.__file__)

    default_config_file = os.path.join(mod_dir, "configs", "defaults.yml")
    default_config = load_yml(default_config_file)

    default_config = default_config

    regex_config_file = os.path.join(mod_dir, "configs", "regex.yml")
    regex_config = load_yml(regex_config_file)

    regex_config = regex_config

    # Parse all the names in the dat file
    all_file_dict = {}
    for f in all_files:

        dir_name = get_directory_name(f)
        full_name = normalize_name(
            f,
            disc_rename=default_config["disc_rename"],
        )
        short_name = get_short_name(
            full_name,
            regex_config=regex_config,
            default_config=default_config,
        )
        region_free_name = get_region_free_name(
            full_name,
            regex_config=regex_config,
            default_config=default_config,
        )

        all_file_dict[f] = {
            "original_name": f,
            "dir_name": dir_name,
            "full_name": full_name,
            "short_name": short_name,
            "region_free_name": region_free_name,
            "matched": False,
        }

    return all_file_dict


def parse_games_from_dat(
    config_file,
    platform,
):
    """Using a config file and platform, parse all the games from a dat

    This will include the dupe parsing and everything if specified in the config
    """

    config = load_yml(config_file)

    parsed_dat_dir = config.get("dirs", {}).get("parsed_dat_dir", None)

    if parsed_dat_dir is None:
        raise ValueError("No parsed_dat_dir found in config file")

    dat_filename = os.path.join(parsed_dat_dir, f"{platform} (dat parsed).json")

    if not os.path.exists(dat_filename):
        raise ValueError("No parsed dat file found")

    all_file_dict = check_regex_parsing(dat_filename)

    finder = GameFinder(
        config_file=config_file,
        platform=platform,
    )
    finder.logger.setLevel(logging.WARNING)

    all_games = finder.run(files=all_file_dict)

    return all_games



import copy
import os
import requests

import romsearch
from ..util import (
    centred_string,
    setup_logger,
    load_yml,
    load_json,
    save_json,
)

ID_CLONE_KEYS = [
    "cloneof",
    "cloneofid",
]


class DupeParser:

    def __init__(
        self,
        platform=None,
        config_file=None,
        config=None,
        default_config=None,
        regex_config=None,
        logger=None,
        log_line_sep="=",
        log_line_length=100,
    ):
        """Tool for figuring out a list of dupes

        Args:
            platform (str, optional): Platform name. Defaults to None, which will throw a ValueError.
            config_file (str, optional): Path to config file. Defaults to None
            config (dict, optional): Configuration dictionary. Defaults to None
            default_config (dict, optional): Default configuration dictionary. Defaults to None
            regex_config (dict, optional): Configuration dictionary for regex search. Defaults to None
            logger (logging.Logger, optional): Logger instance. Defaults to None
            log_line_length (int, optional): Line length of log. Defaults to 100

        TODO:
            - At some point, we might want to consider adding in the retool supersets
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get(
                "log_dir", os.path.join(os.getcwd(), "logs")
            )
            log_level = self.config.get("logger", {}).get("level", "info")
            logger = setup_logger(
                log_level=log_level,
                script_name=f"DupeParser",
                log_dir=log_dir,
                additional_dir=platform,
            )
        self.logger = logger

        # self.use_dat = self.config.get("dupeparser", {}).get("use_dat", True)
        self.use_retool = self.config.get("dupeparser", {}).get("use_retool", True)

        self.parsed_dat_dir = self.config.get("dirs", {}).get("parsed_dat_dir", None)
        if self.use_retool and self.parsed_dat_dir is None:
            raise ValueError("Must specify parsed_dat_dir if using retool files")

        self.dupe_dir = self.config.get("dirs", {}).get("dupe_dir", None)
        if self.dupe_dir is None:
            raise ValueError("dupe_dir should be specified in config file")

        # Pull in platform config that we need
        mod_dir = os.path.dirname(romsearch.__file__)
        retool_config_file = os.path.join(
            mod_dir, "configs", "clonelists", f"retool.yml"
        )
        retool_config = load_yml(retool_config_file)

        self.retool_url = retool_config.get("url", None)
        self.retool_platform_file = retool_config.get(platform, None)

        if default_config is None:
            default_file = os.path.join(mod_dir, "configs", "defaults.yml")
            default_config = load_yml(default_file)
        self.default_config = default_config

        if regex_config is None:
            regex_file = os.path.join(mod_dir, "configs", "regex.yml")
            regex_config = load_yml(regex_file)
        self.regex_config = regex_config

        self.log_line_sep = log_line_sep
        self.log_line_length = log_line_length

    def run(self):
        """Run the dupe parser"""

        if (
            self.retool_platform_file is None or self.retool_url is None
        ) and self.use_retool:
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            self.logger.warning(
                centred_string(
                    "retool config for the platform needs to be present "
                    "if using retool",
                    total_length=self.log_line_length,
                )
            )
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            return False

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")
        self.logger.info(
            centred_string("Running DupeParser", total_length=self.log_line_length)
        )
        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        dupe_dict, retool_dict = self.get_dupe_dict()

        # Save out the dupe dict
        out_file = os.path.join(self.dupe_dir, f"{self.platform} (dupes).json")
        save_json(dupe_dict, out_file)

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        return dupe_dict, retool_dict

    def get_dupe_dict(self):
        """Loop through potentially the retool file to get out dupes"""

        dupe_dict = {}

        # Retool dupes
        retool_dict = None
        if self.use_retool:
            dupe_dict, retool_dict = self.get_retool_dupes(dupe_dict)

        dupe_dict = dict(sorted(dupe_dict.items()))

        return dupe_dict, retool_dict

    def get_retool_dupes(self, dupe_dict=None):
        """Get dupes from the retool curated list"""

        if dupe_dict is None:
            dupe_dict = {}

        retool_dupes = self.get_retool_dupe_dict()
        for retool_dupe in retool_dupes:

            # If we don't have titles or compilations within the dupe dict, skip
            if "titles" not in retool_dupe and "compilations" not in retool_dupe:
                continue

            # Get group name
            group = retool_dupe["group"]

            # Ensure we've not mismatched upper/lowercase
            group_lower = group.lower()
            dupe_dict_keys = [key for key in dupe_dict.keys()]
            dupe_dict_keys_lower = [key.lower() for key in dupe_dict.keys()]

            if group_lower not in dupe_dict_keys_lower:
                final_group = copy.deepcopy(group)
            else:
                final_group_idx = dupe_dict_keys_lower.index(group_lower)
                final_group = dupe_dict_keys[final_group_idx]

            # Pull out individual titles
            if "titles" in retool_dupe:

                if final_group not in dupe_dict:
                    dupe_dict[final_group] = {}

                for title in retool_dupe["titles"]:
                    title_g = title["searchTerm"]
                    name_type = title.get("nameType", None)
                    priority = title.get("priority", 1)
                    filters = title.get("filters", None)

                    dupe_dict[group][title_g] = {
                        "name_type": name_type,
                        "priority": priority,
                        "filters": filters,
                    }

            # Next, check for compilations. If we have them, pull them out and potentially the title position
            if "compilations" in retool_dupe:

                if final_group not in dupe_dict:
                    dupe_dict[final_group] = {}

                for compilation in retool_dupe["compilations"]:
                    comp_g = compilation["searchTerm"]
                    name_type = compilation.get("nameType", None)
                    title_pos = compilation.get("titlePosition", None)
                    priority = compilation.get("priority", 1)
                    filters = compilation.get("filters", None)

                    dupe_dict[group][comp_g] = {
                        "name_type": name_type,
                        "is_compilation": True,
                        "priority": priority,
                        "title_pos": title_pos,
                        "filters": filters,
                    }

        return dupe_dict, retool_dupes

    def download_retool_dupe(
        self,
        out_file=None,
        just_date=False,
    ):
        """Download the retool curated list, optionally just returning the last modified date"""

        retool_url = f"{self.retool_url}/{self.retool_platform_file}"
        with requests.get(retool_url) as r:
            retool_dict = r.json()
            if just_date:
                return retool_dict["description"]["lastUpdated"]
            retool_full_file = r.text

        if out_file is None:
            raise ValueError(
                "Should specify an out_file to save the retool dupe list to"
            )

        with open(out_file, "w", encoding="utf-8") as f:
            f.write(retool_full_file)

        return True

    def get_retool_dupe_dict(self):
        """Pull the retool duplicates out of the clonelist file"""

        if not os.path.exists(self.dupe_dir):
            os.makedirs(self.dupe_dir)

        retool_dupe_file = os.path.join(
            self.parsed_dat_dir, f"{self.platform} (retool).json"
        )
        if not os.path.exists(retool_dupe_file):

            if not os.path.exists(self.parsed_dat_dir):
                os.makedirs(self.parsed_dat_dir)

            self.logger.info(
                centred_string(
                    "No retool dupe file found. Downloading",
                    total_length=self.log_line_length,
                )
            )
            self.download_retool_dupe(retool_dupe_file)

        retool_dupes = load_json(retool_dupe_file)

        # Check if there's a more updated file, if so download it
        local_file_time = retool_dupes["description"]["lastUpdated"]
        remote_file_time = self.download_retool_dupe(just_date=True)

        if not local_file_time == remote_file_time:
            self.logger.info(
                centred_string(
                    "More up-to-date dupe file found. Will download",
                    total_length=self.log_line_length,
                )
            )
            self.download_retool_dupe(retool_dupe_file)

        self.logger.info(
            centred_string(
                f"Using retool clonelist {retool_dupe_file}",
                total_length=self.log_line_length,
            )
        )

        retool_dupes = load_json(retool_dupe_file)
        retool_dupes = retool_dupes["variants"]

        return retool_dupes



from urllib import parse as urlparse

import glob
import os
import re
import shutil
import subprocess
import wget

import romsearch
from ..util import (
    load_yml,
    setup_logger,
    unzip_file,
    centred_string,
    left_aligned_string,
)

ALLOWED_PATCH_METHODS = [
    "xdelta",
    "rompatcher.js",
]


def find_file_by_extensions(file_exts, patch_dir):
    """Find a file by extension

    Args:
        file_exts (list): File extensions to loop over
        patch_dir (str): Patch directory
    """

    potential_files = []

    for file_ext in file_exts:
        potential_file = glob.glob(os.path.join(patch_dir, f"*{file_ext}"))
        potential_files.extend(potential_file)

    # If we have multiple potential files, then error out
    if len(potential_files) > 1:
        raise ValueError(f'Multiple files found: {", ".join(potential_files)}')

    file = potential_files[0]

    return file


class ROMPatcher:

    def __init__(
        self,
        platform,
        config_file=None,
        config=None,
        platform_config=None,
        logger=None,
        log_line_sep="=",
        log_line_length=100,
    ):
        """ROM Patching tool

        There are different ways to patch files based on platforms, so we need
        to keep track of a number of things here

        Args:
            platform (str): Platform name
            config_file (str, optional): path to config file. Defaults to None.
            config (dict, optional): configuration dictionary. Defaults to None.
            platform_config (dict, optional): platform configuration dictionary. Defaults to None.
            logger (logging.Logger, optional): logger. Defaults to None.
            log_line_length (int, optional): Line length of log. Defaults to 100
        """

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get(
                "log_dir", os.path.join(os.getcwd(), "logs")
            )
            log_level = self.config.get("logger", {}).get("level", "info")
            logger = setup_logger(
                log_level=log_level,
                script_name=f"ROMPatcher",
                log_dir=log_dir,
            )
        self.logger = logger

        # Pull in directories
        self.patch_dir = self.config.get("dirs", {}).get("patch_dir", None)
        if self.patch_dir is None:
            raise ValueError("patch_dir needs to be defined in config")

        if not os.path.exists(self.patch_dir):
            os.makedirs(self.patch_dir)

        self.platform = platform

        # Pull in platform config that we need
        mod_dir = os.path.dirname(romsearch.__file__)

        if platform_config is None:
            platform_config_file = os.path.join(
                mod_dir, "configs", "platforms", f"{platform}.yml"
            )
            platform_config = load_yml(platform_config_file)
        self.platform_config = platform_config

        self.log_line_sep = log_line_sep
        self.log_line_length = log_line_length

    def run(
        self,
        file,
        patch_url,
    ):
        """Run the ROMPatcher"""

        filename_no_ext = os.path.splitext(os.path.basename(file))[0]
        patch_dir = str(os.path.join(self.patch_dir, self.platform, filename_no_ext))

        # Clean out and create patch directory
        if os.path.exists(patch_dir):
            shutil.rmtree(patch_dir)
        if not os.path.exists(patch_dir):
            os.makedirs(patch_dir)

        # Move and unzip file, if needed
        self.logger.info(
            centred_string(
                f"Moving {file} to {patch_dir}",
                total_length=self.log_line_length,
            )
        )

        # If we have a zip, unzip that cutie
        if file.endswith(".zip"):
            unzip_file(file, patch_dir)
        else:
            ensure_directory = ""
            if not self.patch_dir.endswith(os.path.sep):
                ensure_directory = os.path.sep
            shutil.copy(file, patch_dir + ensure_directory)

        # Find the unpatched file
        unpatched_file = self.get_unpatched_file(patch_dir=patch_dir)

        # Next up, download the patch file
        patch_file = self.download_patch_file(
            patch_url,
            patch_dir=patch_dir,
        )

        # Now we have everything we need to patch this ROM
        patched_file = self.patch_rom(
            unpatched_file=unpatched_file,
            patch_dir=patch_dir,
            patch_file=patch_file,
        )

        return patched_file

    def get_unpatched_file(
        self,
        patch_dir,
    ):
        """Get the unpatched file from the patch directory

        Args:
            patch_dir (str): Patch directory
        """

        file_exts = self.platform_config.get("file_exts", [])

        # Error if we don't have file extensions defined
        if len(file_exts) == 0:
            raise ValueError(
                "File extensions need to be defined in the platform config file"
            )

        rom_file = find_file_by_extensions(
            file_exts=file_exts,
            patch_dir=patch_dir,
        )

        return rom_file

    def download_patch_file(
        self,
        patch_url,
        patch_dir,
    ):
        """Download a patch file

        Args:
            patch_url (str): URL to patch file
            patch_dir (str): Patch directory
        """

        self.logger.info(
            centred_string(
                f"Downloading patch file: {patch_url}",
                total_length=self.log_line_length,
            )
        )

        # Since the URL can already have the % in, unquote before passing to wget
        patch_file = wget.download(urlparse.unquote(patch_url), out=patch_dir)
        if patch_file.endswith(".zip"):
            unzip_file(patch_file, patch_dir)

        # Find the patch file
        patch_file_exts = self.platform_config.get("patch_file_exts", [])

        # Error if we don't have patch file extensions defined
        if len(patch_file_exts) == 0:
            raise ValueError(
                "Patch file extensions need to be defined in the platform config file"
            )

        patch_file = find_file_by_extensions(
            patch_file_exts,
            patch_dir=patch_dir,
        )

        return patch_file

    def patch_rom(
        self,
        unpatched_file,
        patch_file,
        patch_dir,
    ):
        """Patch a ROM

        Args:
            unpatched_file (str): ROM file to patch
            patch_file (str): Patch file to patch
            patch_dir (str): Patch directory
        """

        # Get the method we're using to patch things
        patch_method = self.platform_config.get("patch_method", None)

        # Error out if the patch method is not defined
        if patch_method is None:
            raise ValueError(
                "Patch method needs to be defined in the platform config file"
            )

        # Build an output file, adding a (ROMPatched) to the bit before the file extension
        unpatch_file_split = os.path.splitext(unpatched_file)
        patched_file = f"{unpatch_file_split[0]} (ROMPatched){unpatch_file_split[1]}"

        if patch_method == "xdelta":
            self.xdelta_patch(
                unpatched_file=unpatched_file,
                patch_file=patch_file,
                out_file=patched_file,
            )
        elif patch_method == "rompatcher.js":

            rompatcher_js_file = (
                f"{unpatch_file_split[0]} (patched){unpatch_file_split[1]}"
            )

            self.rompatcher_js_patch(
                unpatched_file=unpatched_file,
                patch_file=patch_file,
                rompatcher_js_file=rompatcher_js_file,
                out_file=patched_file,
                patch_dir=patch_dir,
            )
        else:
            raise ValueError(
                f"Patch method needs to be one of {', '.join(ALLOWED_PATCH_METHODS)}, not {patch_method}"
            )

        self.logger.info(
            centred_string(
                f"Patching complete!",
                total_length=self.log_line_length,
            )
        )

        return patched_file

    def xdelta_patch(
        self,
        unpatched_file,
        patch_file,
        out_file,
    ):
        """Patch using xdelta

        Args:
            unpatched_file (str): ROM file to patch
            patch_file (str): Patch file to patch
            out_file (str): Path for output file
        """

        xdelta_path = self.config.get("rompatcher", {}).get("xdelta_path", None)

        if xdelta_path is None:
            raise ValueError("Path to xdelta needs to be defined in user config")

        if not os.path.exists(xdelta_path):
            raise ValueError("xdelta path not found")

        cmd = f'{xdelta_path} -d -s "{unpatched_file}" "{patch_file}" "{out_file}"'

        self.logger.info(
            centred_string(
                f"Patching file with xdelta:",
                total_length=self.log_line_length,
            )
        )
        self.logger.info(
            left_aligned_string(
                f"-> Unpatched file: {os.path.basename(unpatched_file)}",
                total_length=self.log_line_length,
            )
        )
        self.logger.info(
            left_aligned_string(
                f"-> Patch file: {os.path.basename(patch_file)}",
                total_length=self.log_line_length,
            )
        )
        self.logger.info(
            left_aligned_string(
                f"-> Output file: {os.path.basename(out_file)}",
                total_length=self.log_line_length,
            )
        )

        os.system(cmd)

        return True

    def rompatcher_js_patch(
        self,
        unpatched_file,
        patch_file,
        rompatcher_js_file,
        out_file,
        patch_dir,
    ):
        """Patch using RomPatcher.js

        Args:
            unpatched_file (str): ROM file to patch
            patch_file (str): Patch file to patch
            rompatcher_js_file (str): Filename that RomPatcher.js will output
            out_file (str): Path for output file
            patch_dir (str): Patch directory
        """

        rompatcher_js_path = self.config.get("rompatcher", {}).get(
            "rompatcher_js_path", None
        )

        if rompatcher_js_path is None:
            raise ValueError("Path to RomPatcher.js needs to be defined in user config")

        if not os.path.exists(rompatcher_js_path):
            raise ValueError("RomPatcher.js path not found")

        cmd = f'node {rompatcher_js_path} patch "{unpatched_file}" "{patch_file}"'

        self.logger.info(
            centred_string(
                f"Patching file with RomPatcher.js:",
                total_length=self.log_line_length,
            )
        )
        self.logger.info(
            left_aligned_string(
                f"-> Unpatched file: {os.path.basename(unpatched_file)}",
                total_length=self.log_line_length,
            )
        )
        self.logger.info(
            left_aligned_string(
                f"-> Patch file: {os.path.basename(patch_file)}",
                total_length=self.log_line_length,
            )
        )

        self.logger.info(
            left_aligned_string(
                f"-> RomPatcher.js file: {os.path.basename(rompatcher_js_file)}",
                total_length=self.log_line_length,
            )
        )

        # Change to patch directory so file ends up in a sensible spot
        orig_dir = os.getcwd()
        os.chdir(patch_dir)

        with subprocess.Popen(
            cmd, text=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        ) as process:
            for line in process.stdout:

                # Replace any potential tabs in the line, strip whitespace and skip newline at the end
                line = re.sub("\s+", " ", line[:-1])
                line = line.lstrip().rstrip()

                if len(line) == 0:
                    continue

                # Log each line of the output using the provided logger
                self.logger.info(
                    centred_string(line, total_length=self.log_line_length)
                )

        # Return to working directory
        os.chdir(orig_dir)

        self.logger.info(
            left_aligned_string(
                f"-> Renaming file to: {os.path.basename(out_file)}",
                total_length=self.log_line_length,
            )
        )

        shutil.copy(rompatcher_js_file, out_file)
        os.remove(rompatcher_js_file)

        return True



import copy
import os
import requests
import time
from datetime import datetime

import romsearch
from ..util import (
    centred_string,
    load_yml,
    setup_logger,
    load_json,
    save_json,
    get_file_time,
)

RA_URL = "https://retroachievements.org/API"


class RAHasher:

    def __init__(
        self,
        platform,
        config_file=None,
        config=None,
        platform_config=None,
        logger=None,
        log_line_sep="=",
        log_line_length=100,
    ):
        """Get supported ROM files and hashes for RetroAchievements

        This works per-platform, so must be specified here

        Args:
            platform (str): Platform name
            config_file (str, optional): Path to config file. Defaults to None.
            config (dict, optional): Configuration dictionary. Defaults to None.
            platform_config (dict, optional): Platform configuration dictionary. Defaults to None
            logger (logging.Logger, optional): Logger instance. Defaults to None.
            log_line_length (int, optional): Line length of log. Defaults to 100
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        # Read in the specific platform configuration
        mod_dir = os.path.dirname(romsearch.__file__)

        if platform_config is None:
            platform_config_file = os.path.join(
                mod_dir, "configs", "platforms", f"{platform}.yml"
            )
            platform_config = load_yml(platform_config_file)
        self.platform_config = platform_config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get(
                "log_dir", os.path.join(os.getcwd(), "logs")
            )
            log_level = self.config.get("logger", {}).get("level", "info")
            logger = setup_logger(
                log_level=log_level,
                script_name=f"RAHasher",
                log_dir=log_dir,
            )
        self.logger = logger

        self.ra_hash_dir = self.config.get("dirs", {}).get("ra_hash_dir", None)

        self.username = self.config.get("rahasher", {}).get("username", None)
        self.api_key = self.config.get("rahasher", {}).get("api_key", None)

        cache_period = self.config.get("rahasher", {}).get("cache_period", 30)
        if isinstance(cache_period, str):
            cache_period = float(cache_period)
        self.cache_period = cache_period

        self.log_line_sep = log_line_sep
        self.log_line_length = log_line_length

    def run(self):
        """Run the RA hasher"""

        run_rahasher = True
        ra_game_info = None

        if self.ra_hash_dir is None:
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            self.logger.warning(
                centred_string(
                    "No parsed_dat_dir defined in config file",
                    total_length=self.log_line_length,
                )
            )
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            run_rahasher = False
        if "ra_id" not in self.platform_config:
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            self.logger.warning(
                centred_string(
                    "RA ID not defined in platform config file",
                    total_length=self.log_line_length,
                )
            )
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            run_rahasher = False
        if self.username is None:
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            self.logger.warning(
                centred_string(
                    "RA username needs to be defined in config file",
                    total_length=self.log_line_length,
                )
            )
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            run_rahasher = False
        if self.api_key is None:
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            self.logger.warning(
                centred_string(
                    "RA API key needs to be defined in config file",
                    total_length=self.log_line_length,
                )
            )
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            run_rahasher = False

        if run_rahasher:

            if not os.path.exists(self.ra_hash_dir):
                os.makedirs(self.ra_hash_dir)

            ra_game_info = self.run_rahasher()

        return ra_game_info

    def run_rahasher(self):
        """The main meat of running the RA hasher"""

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")
        self.logger.info(
            centred_string(
                f"Running RAHasher for {self.platform}",
                total_length=self.log_line_length,
            )
        )
        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        ra_game_info = self.get_ra_game_info()

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        return ra_game_info

    def get_ra_game_info(self):
        """Download or read in RA game info."""

        game_list_file = os.path.join(
            self.ra_hash_dir, f"{self.platform} GameList.json"
        )
        game_list_modtime = get_file_time(
            game_list_file,
            return_as_str=False,
        )

        cur_time = datetime.now()
        diff = cur_time - game_list_modtime

        # If we're still in the cache period, don't overwrite
        self.logger.info(
            centred_string(
                f"Last GameList cache {diff.days} days ago",
                total_length=self.log_line_length,
            )
        )
        if diff.days > self.cache_period:
            self.logger.info(
                centred_string(
                    f"Downloading latest GameList to {game_list_file}",
                    total_length=self.log_line_length,
                )
            )
            self.download_ra_game_list(out_file=game_list_file)

        # Now pull out the game info
        game_info_file = os.path.join(
            self.ra_hash_dir,
            f"{self.platform}.json",
        )
        game_info_modtime = get_file_time(
            game_info_file,
            return_as_str=False,
        )
        cur_time = datetime.now()
        diff = cur_time - game_info_modtime

        self.logger.info(
            centred_string(
                f"Last game info cache {diff.days} days ago",
                total_length=self.log_line_length,
            )
        )
        # If we're still in the cache period, don't overwrite
        if diff.days > self.cache_period:
            self.logger.info(
                centred_string(
                    f"Pulling game info out to {game_info_file}",
                    total_length=self.log_line_length,
                )
            )
            self.download_ra_game_list(out_file=game_list_file)

            ra_game_info = self.format_game_list(game_list_file)

            self.save_ra_game_info(
                ra_game_info,
                game_info_file,
            )

        else:
            self.logger.info(
                centred_string(
                    f"Loading existing game info file",
                    total_length=self.log_line_length,
                )
            )
            ra_game_info = load_json(game_info_file)

        return ra_game_info

    def download_ra_game_list(
        self,
        out_file,
    ):
        """Download the RA GameList using the API

        Args:
            out_file (string): output file name.
        """

        console_id = self.platform_config["ra_id"]

        url = f"{RA_URL}/API_GetGameList.php?z={self.username}&y={self.api_key}&i={console_id}&h=1&f=1"
        resp = requests.get(url=url)
        data = resp.json()

        save_json(data, out_file)

        return True

    def format_game_list(
        self,
        in_file,
        sleep_time=0.5,
    ):
        """Format the GameList neatly

        Args:
            in_file (str): Input GameList file
            sleep_time (float): Sleep time for API queries. Defaults to 0.5
        """

        data = load_json(in_file)

        ra_game_info = {}

        # To not totally clutter the terminal, overwrite each line in the loop
        orig_terminator = self.logger.handlers[-1].terminator
        self.logger.handlers[-1].terminator = ""

        tot = len(data)
        for i, d in enumerate(data):

            # RA uses "Title: Subtitle" logic, while No-Intro/Redump do not
            clean_title = d["Title"].replace(":", " -")

            # Format this string neatly
            self.logger.info(
                centred_string(
                    f"{i+1}/{tot}: {clean_title}",
                    total_length=self.log_line_length,
                    str_prefix="\rINFO: ",
                )
            )

            # Use ID to get all the useful hash info
            game_hashes = self.get_game_hash(d["ID"])
            d["Hashes"] = copy.deepcopy(game_hashes)

            ra_game_info[clean_title] = copy.deepcopy(d)

            # To avoid overloading the API, add a pause here
            time.sleep(sleep_time)

        # Set the logger back to what it was,
        # and avoid any last minute formatting errors
        self.logger.handlers[-1].terminator = orig_terminator
        self.logger.info("")

        return ra_game_info

    def get_game_hash(
        self,
        game_id,
    ):
        """Using a game ID, get list of the game hashes"""

        url = f"{RA_URL}/API_GetGameHashes.php?i={game_id}&y={self.api_key}&z={self.username}"
        resp = requests.get(url=url)
        game_hashes = resp.json()

        game_hashes = game_hashes["Results"]

        return game_hashes

    def save_ra_game_info(
        self,
        ra_game_info,
        out_file,
    ):
        """Save out the RA game info to a file

        Args:
            ra_game_info (dict): RA game info
            out_file (string): output file name
        """

        if not os.path.exists(self.ra_hash_dir):
            os.makedirs(self.ra_hash_dir)

        self.logger.info(
            centred_string(
                f"Saving full game info to {out_file}",
                total_length=self.log_line_length,
            )
        )

        save_json(
            ra_game_info,
            out_file,
        )

        return True



from collections import Counter

import copy
import numpy as np
import os
from packaging import version

import romsearch
from ..util import (
    setup_logger,
    centred_string,
    left_aligned_string,
    load_yml,
)

DAT_FILTERS = [
    "add-ons",
    "applications",
    "audio",
    "bad_dumps",
    "console",
    "bonus_discs",
    "coverdiscs",
    "demos",
    "educational",
    "games",
    "manuals",
    "multimedia",
    "pirate",
    "preproduction",
    "promotional",
    "unlicensed",
    "video",
]


def get_sanitized_version(ver):
    """Get a sanitized version for potentially weird versioning

    Args:
        ver: version to sanitize
    """

    # If we don't have a version, assume v0
    if ver.lower() in [""]:
        ver = "v0"

    # If we just have "rev", assume a v1
    if ver.lower() in ["rev"]:
        ver = "v1"

    # If we have lettered versions, convert these to numbers
    if ver[1:].isalpha():
        ver = f"v{ord(ver[1:])}"

    # If we have a v.something, replace that here so versions can be parsed
    if ver[:2] == "v.":
        ver = f"v{ver[2:]}"

    # If we have a letter at the end, replace with an ordinal
    if ver[-1].isalpha():
        ver = f"{ver[:-1]}.{ord(ver[-1])}"

    return ver


def argsort(seq):
    return sorted(range(len(seq)), key=seq.__getitem__)


def remove_rom_dict_entries(
    rom_dict,
    key,
    remove_type="bool",
    bool_remove=True,
    list_preferences=None,
):
    """Exclude entries from the game dict based on various rules"""

    f_to_exclude = []
    for f in rom_dict:

        # Bool type, we can filter either on Trues or Falses
        if remove_type == "bool":

            key_val = rom_dict[f].get(key, None)
            if key_val is None:
                continue

            if bool_remove:
                if key_val:
                    f_to_exclude.append(f)
            else:
                if not key_val:
                    f_to_exclude.append(f)

        elif remove_type == "list":

            if list_preferences is None:
                raise ValueError("list_preferences not specified")

            # If there's no information in there, assume we're OK
            if len(rom_dict[f][key]) == 0:
                continue

            found = False
            for val in rom_dict[f][key]:
                if val in list_preferences:
                    found = True
            if not found:
                f_to_exclude.append(f)

        else:
            raise ValueError("remove_type should be one of bool, list")

    f_to_exclude = np.unique(f_to_exclude)
    f_to_exclude = [str(f) for f in f_to_exclude]
    for f in f_to_exclude:
        rom_dict[f]["excluded"] = True
        rom_dict[f]["excluded_reason"].append(key)

    return rom_dict


def add_versioned_score(files, rom_dict, key):
    """Get an order for versioned strings"""

    # Ensure we have a version here. If blank, set to v0
    rom_dict = copy.deepcopy(rom_dict)
    for f in rom_dict:
        rom_dict[f][key] = get_sanitized_version(rom_dict[f][key])

    versions = [version.parse(rom_dict[f][key]) for f in files]
    versions_sorted = np.unique(sorted(versions))

    file_scores_version = np.zeros(len(files))
    for i, v in enumerate(versions_sorted):
        v_idx = np.where(np.asarray(versions) == v)[0]
        file_scores_version[v_idx] += i

    return file_scores_version


def add_ordered_score(files, rom_dict, priorities, score_key):
    """Add an ordered score, include the priority of the order

    Args:
        files (list): List of files to score
        rom_dict (dict): Dictionary of ROMs
        priorities (list): List of values in priority order
        score_key (str): Corresponding score key for the priorities in the rom_dict
    """

    score_dict = {}

    # Go backwards so the first entry is the highest priority
    for i, prio in enumerate(priorities[::-1]):
        score_dict[prio] = i + 1

    scores = np.zeros_like(files, dtype=int)
    for i, f in enumerate(files):
        for key in rom_dict[f][score_key]:
            if key in score_dict:
                scores[i] += score_dict[key]

    return scores


def filter_by_list(
    rom_dict,
    key,
    key_prefs,
):
    """Find file with highest value in given list. If there are multiple matches, find the most updated one"""

    roms = []

    found_key = False
    for key_pref in key_prefs:

        if found_key:
            continue

        for val in rom_dict:

            # Only count this as "found" if we haven't already excluded
            if key_pref in rom_dict[val][key] and not rom_dict[val]["excluded"]:
                roms.append(val)
                found_key = True

    keys_to_exclude = []
    for f in rom_dict:
        if f not in roms:
            keys_to_exclude.append(f)

    for e in keys_to_exclude:

        # Only exclude by preference if we haven't already excluded absolutely
        if key not in rom_dict[e]["excluded_reason"]:
            rom_dict[e]["excluded"] = True
            rom_dict[e]["excluded_reason"].append(f"{key}_preference")

    return rom_dict


def filter_compilations(
    rom_dict,
):
    """Filter out compilations if we have them

    Args:
        rom_dict (dict): Dictionary of ROMs
    """

    compilations = []
    non_compilations = []

    for r in rom_dict:
        if not rom_dict[r]["excluded"]:
            if rom_dict[r].get("is_compilation", False):
                compilations.append(r)
            else:
                non_compilations.append(r)

    # Only remove compilations if we have singular titles
    if len(non_compilations) > 0:
        for comp in compilations:
            rom_dict[comp]["excluded"] = True
            rom_dict[comp]["excluded_reason"].append("is_compilation")

    return rom_dict


class ROMChooser:

    def __init__(
        self,
        platform,
        game,
        config_file=None,
        config=None,
        default_config=None,
        regex_config=None,
        logger=None,
        log_line_sep="=",
        log_line_length=100,
    ):
        """ROM choose tool

        This works per-game, per-platform, so must be specified here

        Args:
            platform (str): Platform name
            game (str): Game name
            config_file (str, optional): Path to config file. Defaults to None.
            config (dict, optional): Configuration dictionary. Defaults to None.
            default_config (dict, optional): Default configuration dictionary. Defaults to None.
            regex_config (dict, optional): Configuration dictionary. Defaults to None.
            logger (logging.Logger, optional): Logger instance. Defaults to None.
            log_line_length (int, optional): Line length of log. Defaults to 100
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if game is None:
            raise ValueError("game must be specified")
        self.game = game

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get(
                "log_dir", os.path.join(os.getcwd(), "logs")
            )
            logger_add_dir = str(os.path.join(platform, game))
            log_level = self.config.get("logger", {}).get("level", "info")
            logger = setup_logger(
                log_level=log_level,
                script_name=f"ROMChooser",
                log_dir=log_dir,
                additional_dir=logger_add_dir,
            )
        self.logger = logger

        mod_dir = os.path.dirname(romsearch.__file__)

        if default_config is None:
            default_file = os.path.join(mod_dir, "configs", "defaults.yml")
            default_config = load_yml(default_file)
        self.default_config = default_config

        if regex_config is None:
            regex_file = os.path.join(mod_dir, "configs", "regex.yml")
            regex_config = load_yml(regex_file)
        self.regex_config = regex_config

        # Region preference (usually set USA for retroachievements, can also be a list to fall back to)
        region_preferences = self.config.get(
            "region_preferences", self.default_config["default_region"]
        )
        if isinstance(region_preferences, str):
            region_preferences = [region_preferences]

        for region_pref in region_preferences:
            if region_pref not in self.default_config["regions"]:
                raise ValueError(
                    f"Regions should be any of {self.default_config['regions']}, not {region_pref}"
                )

        self.region_preferences = region_preferences

        # Language preference (usually set En, can also be a list to fall back to)
        language_preferences = self.config.get(
            "language_preferences", self.default_config["default_language"]
        )
        if isinstance(language_preferences, str):
            language_preferences = [language_preferences]

        for language_pref in language_preferences:
            if language_pref not in self.default_config["languages"]:
                raise ValueError(
                    f"Regions should be any of {self.default_config['languages']}, not {language_pref}"
                )

        self.language_preferences = language_preferences

        # Various filters. First are the boolean ones
        dat_filters = self.config.get("romchooser", {}).get(
            "dat_filters", "all_but_games"
        )
        if "all" in dat_filters:
            all_dat_filters = copy.deepcopy(DAT_FILTERS)
            if "all_but" in dat_filters:
                filter_to_remove = dat_filters.split("all_but_")[-1]
                all_dat_filters.remove(filter_to_remove)
            dat_filters = copy.deepcopy(all_dat_filters)

        if isinstance(dat_filters, str):
            dat_filters = [dat_filters]
        self.dat_filters = dat_filters

        self.filter_regions = self.config.get("romchooser", {}).get(
            "filter_regions", True
        )
        self.filter_languages = self.config.get("romchooser", {}).get(
            "filter_languages", True
        )
        self.use_best_version = self.config.get("romchooser", {}).get(
            "use_best_version", True
        )

        self.dry_run = self.config.get("romchooser", {}).get("dry_run", False)

        self.log_line_sep = log_line_sep
        self.log_line_length = log_line_length

    def run(self, rom_dict):
        """Run the ROM chooser"""

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")
        self.logger.info(
            centred_string(
                f"Running ROMChooser for {self.game}", total_length=self.log_line_length
            )
        )
        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        rom_dict = self.run_chooser(rom_dict)

        self.print_summary(rom_dict)

        # Filter out excluded ROMs before we return
        rom_dict = {
            key: rom_dict[key] for key in rom_dict if not rom_dict[key]["excluded"]
        }

        self.logger.info(f"{self.log_line_sep * self.log_line_length}")

        return rom_dict

    def run_chooser(self, rom_dict):
        """Make a ROM choice based on various factors

        This chooser works in this order:

        * Filter out dat categories we don't want (e.g. demos, betas)
        * Filter out ROMs that don't have any languages in the user preferences
        * Filter out ROMs that don't have any regions in the user preferences

        For the ROMs left, we then choose a best one, using a scoring system with
        this priority:

        * Achievements
        * Regions
        * Languages
        * Budget editions
        * Versions and revisions
        * Improved versions

        We also demote ROMs, with this priority (most to least demoted):

        * Retool priority
        * Modern versions
        * Alternate versions
        * Demoted versions

        Args:
            rom_dict (dict): Dictionary of ROMs to choose between
        """

        # Add in whether these are excluded or not, and why
        for r in rom_dict:
            rom_dict[r]["excluded"] = False
            rom_dict[r]["excluded_reason"] = []

        for f in self.dat_filters:
            self.logger.debug(
                left_aligned_string(f"Filtering {f}", total_length=self.log_line_length)
            )
            if f in DAT_FILTERS:
                rom_dict = remove_rom_dict_entries(
                    rom_dict,
                    f,
                    remove_type="bool",
                    bool_remove=True,
                )
            else:
                raise ValueError(f"Unknown filter type {f}")

        # Language
        if self.filter_languages:
            self.logger.debug(
                left_aligned_string(
                    f"Filtering languages", total_length=self.log_line_length
                )
            )
            rom_dict = remove_rom_dict_entries(
                rom_dict,
                "languages",
                remove_type="list",
                list_preferences=self.language_preferences,
            )

        # Regions
        if self.filter_regions:
            self.logger.debug(
                left_aligned_string(
                    f"Filtering regions", total_length=self.log_line_length
                )
            )
            rom_dict = remove_rom_dict_entries(
                rom_dict,
                "regions",
                remove_type="list",
                list_preferences=self.region_preferences,
            )

        # If we have a split between singular titles and compilations, sort that out here
        self.logger.debug(
            left_aligned_string(
                f"Potentially filtering compilations", total_length=self.log_line_length
            )
        )
        rom_dict = filter_compilations(
            rom_dict,
        )

        # Remove versions we potentially don't want around
        if self.use_best_version:
            self.logger.debug(
                left_aligned_string(
                    f"Getting best version", total_length=self.log_line_length
                )
            )
            rom_dict = self.get_best_rom(
                rom_dict,
            )

        return rom_dict

    def print_summary(
        self,
        rom_dict,
    ):
        """Log out a nice summary of what ROM has been chosen here

        Args:
            rom_dict (dict): the ROM dictionary to summarize
        """

        if len(rom_dict) == 0:
            # Just say nothing found
            self.logger.info(
                centred_string("No ROMs found", total_length=self.log_line_length)
            )
        else:

            # Start with found ROMs, then excluded ROMs and reasons
            if np.sum([not rom_dict[r]["excluded"] for r in rom_dict]) > 0:
                self.logger.info(
                    centred_string("Included ROMs:", total_length=self.log_line_length)
                )

            for r in rom_dict:

                # Don't include the excluded ones here
                if rom_dict[r]["excluded"]:
                    continue

                self.logger.info(
                    left_aligned_string(f"-> {r}", total_length=self.log_line_length)
                )
                if rom_dict[r]["has_cheevos"]:
                    self.logger.info(
                        left_aligned_string(
                            f"--> Has RetroAchievements",
                            total_length=self.log_line_length,
                        )
                    )
                if rom_dict[r]["patch_file"] != "":
                    self.logger.info(
                        left_aligned_string(
                            f"--> Patch URL: {rom_dict[r]['patch_file']}",
                            total_length=self.log_line_length,
                        )
                    )

            if (
                np.sum([rom_dict[r]["excluded"] for r in rom_dict]) > 0
                and np.sum([not rom_dict[r]["excluded"] for r in rom_dict]) > 0
            ):
                self.logger.info(f"{'-' * self.log_line_length}")

            if np.sum([rom_dict[r]["excluded"] for r in rom_dict]) > 0:
                self.logger.info(
                    centred_string("Excluded ROMs:", total_length=self.log_line_length)
                )

            for r in rom_dict:

                # Don't include the excluded ones here
                if not rom_dict[r]["excluded"]:
                    continue

                # Make the exclusion reason more human-readable
                exclusion_reason = rom_dict[r]["excluded_reason"]
                exclusion_reason = [
                    e.capitalize().replace("_", " ") for e in exclusion_reason
                ]

                self.logger.info(
                    left_aligned_string(f"-> {r}", total_length=self.log_line_length)
                )
                self.logger.info(
                    left_aligned_string(
                        f"--> Reason(s): {', '.join(exclusion_reason)}",
                        total_length=self.log_line_length,
                    )
                )

        return True

    def get_best_roms(
        self,
        files,
        rom_dict,
    ):
        """Get the best ROM(s) from a list, using a scoring system"""

        # Positive scores
        improved_version_score = 1e2
        version_score = 1e4
        revision_score = 1e6
        budget_edition_score = 1e8
        language_score = 1e10
        region_score = 1e12
        cheevo_score = 1e14

        # Negative scores
        compilation_score = -1
        demoted_version_score = -1e2
        alternate_version_score = -1e2
        modern_version_score = -1e4
        priority_score = -1e6

        file_scores = np.zeros(len(files))

        # Just stepping through the scores in order.

        # Positive scores

        # Improved version
        file_scores += improved_version_score * np.array(
            [int(rom_dict[f]["improved_version"]) for f in files]
        )

        # Version numbering, which needs to be parsed. We edit these to only add a little each time
        version_score_to_add = add_versioned_score(files, rom_dict, "version")
        file_scores += version_score * (1 + (version_score_to_add - 1) / 100)

        # Revision numbering, again parsed
        revision_score_to_add = add_versioned_score(files, rom_dict, "revision")
        file_scores += revision_score * (1 + (revision_score_to_add - 1) / 100)

        # Budget edition
        file_scores += budget_edition_score * np.array(
            [int(rom_dict[f]["budget_edition"]) for f in files]
        )

        # Language priorities
        language_score_to_add = add_ordered_score(
            files, rom_dict, priorities=self.language_preferences, score_key="languages"
        )
        file_scores += language_score * (1 + (language_score_to_add - 1) / 100)

        # Regions priorities
        region_score_to_add = add_ordered_score(
            files,
            rom_dict,
            priorities=self.region_preferences,
            score_key="regions",
        )
        file_scores += region_score * (1 + (region_score_to_add - 1) / 100)

        # Achievement hashes
        file_scores += cheevo_score * np.array(
            [int(rom_dict[f]["has_cheevos"]) for f in files]
        )

        # Negative scores

        # Compilation score
        file_scores += compilation_score * np.array(
            [rom_dict[f].get("is_compilation", False) for f in files]
        )

        # Demoted version
        file_scores += demoted_version_score * np.array(
            [int(rom_dict[f]["demoted_version"]) for f in files]
        )

        # Alternate version
        file_scores += alternate_version_score * np.array(
            [int(rom_dict[f]["alternate"]) for f in files]
        )

        # Modern version
        file_scores += modern_version_score * np.array(
            [int(rom_dict[f]["modern_version"]) for f in files]
        )

        # Priority scoring. We subtract 1 so that the highest priority has no change
        file_scores += priority_score * (
            np.array([int(rom_dict[f]["priority"]) for f in files]) - 1
        )

        files_idx = np.where(file_scores == np.nanmax(file_scores))[0]
        files = np.asarray(files)[files_idx]

        return files

    def get_best_rom(
        self,
        rom_dict,
    ):
        """Get an overall best ROM using a scoring system"""

        roms = []

        for key in rom_dict:
            if not rom_dict[key]["excluded"]:
                roms.append(key)

        if len(roms) > 1:
            roms = self.get_best_roms(roms, rom_dict)

            keys_to_exclude = []
            for f in rom_dict:
                if f not in roms:
                    keys_to_exclude.append(f)

            for key in keys_to_exclude:
                # Only exclude here if there isn't already another reason
                if len(rom_dict[key]["excluded_reason"]) == 0:
                    rom_dict[key]["excluded"] = True
                    rom_dict[key]["excluded_reason"].append(f"not_highest_score")

        return rom_dict



from importlib.metadata import version

# Get the version
__version__ = version(__name__)

from .modules import (
    DATParser,
    DupeParser,
    GameFinder,
    RAHasher,
    ROMDownloader,
    ROMChooser,
    ROMCleaner,
    ROMMover,
    ROMParser,
    ROMPatcher,
    ROMSearch,
)

__all__ = [
    "DATParser",
    "DupeParser",
    "GameFinder",
    "RAHasher",
    "ROMDownloader",
    "ROMChooser",
    "ROMCleaner",
    "ROMMover",
    "ROMParser",
    "ROMPatcher",
    "ROMSearch",
]



import copy
import os
import re
from packaging.version import Version

import romsearch
from ..util import (
    centred_string,
    left_aligned_string,
    setup_logger,
    get_file_time,
    load_yml,
    load_json,
    match_retool_search_terms,
)

DICT_DEFAULT_VALS = {"bool": False, "str": "", "list": []}
USE_TITLE_POS = [
    "languages",
]


def find_pattern(regex, search_str, group_number=0):
    """
    Take a regex pattern and find potential matches within a search string
    """
    regex_search_str = None

    regex_search = re.search(regex, search_str)
    if regex_search:
        regex_search_str = regex_search.group(group_number)

    return regex_search_str


def get_pattern_val(
    regex,
    tag,
    regex_type,
    pattern_mappings=None,
    title_pos=None,
    use_title_pos=False,
):
    """Get values out from a regex pattern, optionally mapping back to something more readable for lists

    Args:
        regex: Regex pattern
        tag: Found tag
        regex_type: Regex pattern type. Can be str, bool, list
        pattern_mappings: Mapping from regex pattern to more readable values
        title_pos: Position of title for compilations. Defaults to None
        use_title_pos: Use title_pos? Defaults to False
    """

    pattern_string = find_pattern(regex, tag)

    if pattern_string is not None:
        pattern_string = pattern_string.strip("()")

        # Split out to the specific languages, but only if they're marked correctly
        if title_pos is not None and use_title_pos and "+" in pattern_string:
            pattern_string = pattern_string.split("+")[title_pos - 1]

        if regex_type == "bool":
            pattern_val = True
        elif regex_type == "str":
            pattern_val = pattern_string
        elif regex_type == "list":

            if pattern_mappings is not None:
                parsed_pattern_string = []
                # Match to pattern mappings
                for p in pattern_mappings:
                    if re.search(pattern_mappings[p], pattern_string):
                        parsed_pattern_string.append(p)
            else:
                # Split, and remove and trailing whitespace
                parsed_pattern_string = pattern_string.split(",")
                parsed_pattern_string = [s.strip() for s in parsed_pattern_string]
            pattern_val = parsed_pattern_string
        else:
            raise ValueError("regex_type should be one of 'bool', 'str', or 'list'")

    else:
        pattern_val = None

    return pattern_val


def is_ra_subset(name):
    """Check if a name is a RetroAchievements subset

    Args:
        name (str): Name to check
    """

    match_pattern = "\\[Subset.*\\]"

    match = find_pattern(match_pattern, name)
    is_subset = False
    if match is not None:
        is_subset = True

    return is_subset


def check_match(i, j, checks_passed=None):
    """Check if two bools/strings/lists match

    For lists, we simply check if there's any subset that matches

    Args:
        i: Input 1
        j: Input 2
        checks_passed: If not None, will inherit this as initial start.
            Else, will default to True
    """

    if checks_passed is None:
        checks_passed = True
    if not isinstance(checks_passed, bool):
        raise ValueError("checks_passed should be a boolean value")

    # If we have a bool or string, then they should match
    if isinstance(i, bool) or isinstance(i, str):
        if not i == j:
            checks_passed = False

    # If a list, then check there's at least some overlap
    elif isinstance(i, list):
        s_i = set(i)
        s_j = set(j)
        s_k = s_i.intersection(s_j)

        if len(s_k) == 0:
            checks_passed = False

    else:
        t = type(i)
        raise ValueError(f"Do not know how to check against type {t}")

    return checks_passed


class ROMParser:

    def __init__(
        self,
        platform,
        game,
        dat=None,
        retool=None,
        ra_hashes=None,
        config_file=None,
        config=None,
        platform_config=None,
        default_config=None,
        regex_config=None,
        logger=None,
        log_line_sep="=",
        log_line_length=100,
    ):
        """ROM parser tool

        This works per-game, per-platform, so must be specified here

        Args:
            platform (str): Platform name
            game (str): Game name
            dat (dict): Parsed dat dictionary. Defaults to None, which will try to load the dat file if it exists
            retool (dict): Retool dictionary. Defaults to None, which will try to load the file if it exists
            ra_hashes (dict): RA hash dictionary. Defaults to None, which will try to load the file if it exists
            config_file (str, optional): path to config file. Defaults to None.
            config (dict, optional): configuration dictionary. Defaults to None.
            platform_config (dict, optional): platform configuration dictionary. Defaults to None.
            default_config (dict, optional): default configuration dictionary. Defaults to None.
            regex_config (dict, optional): regex configuration dictionary. Defaults to None.
            logger (logging.Logger, optional): logger instance. Defaults to None.
            log_line_length (int, optional): Line length of log. Defaults to 100

        TODO:
            For the RetroAchievements, there are hacks and unlicensed stuff that seems to work differently
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        self.game = game

        if logger is None:
            log_dir = self.config.get("dirs", {}).get(
                "log_dir", os.path.join(os.getcwd(), "logs")
            )
            logger_add_dir = str(os.path.join(platform, game))
            log_level = self.config.get("logger", {}).get("level", "info")
            logger = setup_logger(
                log_level=log_level,
                script_name=f"ROMParser",
                log_dir=log_dir,
                additional_dir=logger_add_dir,
            )
        self.logger = logger

        mod_dir = os.path.dirname(romsearch.__file__)

        if default_config is None:
            default_file = os.path.join(mod_dir, "configs", "defaults.yml")
            default_config = load_yml(default_file)
        self.default_config = default_config

        self.ra_file_exts = self.default_config.get("ra_file_exts", [])
        self.ra_labels = self.default_config.get("ra_labels", [])
        self.ra_patch_checks = self.default_config.get("ra_patch_checks", [])

        if regex_config is None:
            regex_file = os.path.join(mod_dir, "configs", "regex.yml")
            regex_config = load_yml(regex_file)
        self.regex_config = regex_config

        if platform_config is None:
            platform_config_file = os.path.join(
                mod_dir, "configs", "platforms", f"{platform}.yml"
            )
            platform_config = load_yml(platform_config_file)
        self.platform_config = platform_config

        self.raw_dir = self.config.get("dirs", {}).get("raw_dir", None)
        if not self.raw_dir:
            raise ValueError("raw_dir must be specified in config.yml")

        self.use_dat = self.config.get("romparser", {}).get("use_dat", True)
        self.use_retool = self.config.get("romparser", {}).get("use_retool", True)
        self.use_ra_hashes = self.config.get("romparser", {}).get(
            "use_ra_hashes", False
        )
        self.use_filename = self.config.get("romparser", {}).get("use_filename", True)
        self.dry_run = self.config.get("romparser", {}).get("dry_run", False)

        # If we're using the dat file, pull it out here
        self.dat = dat
        if self.use_dat and self.dat is None:
            dat_dir = self.config.get("dirs", {}).get("parsed_dat_dir", None)
            if dat_dir is None:
                raise ValueError("parsed_dat_dir must be specified in config.yml")
            dat_file = os.path.join(dat_dir, f"{platform} (dat parsed).json")
            if os.path.exists(dat_file):
                self.dat = load_json(dat_file)

        # If we're using the retool file, pull it out here
        self.retool = retool
        if self.use_retool and self.retool is None:
            dat_dir = self.config.get("dirs", {}).get("parsed_dat_dir", None)
            if dat_dir is None:
                raise ValueError("parsed_dat_dir must be specified in config.yml")
            retool_file = os.path.join(dat_dir, f"{platform} (retool).json")
            if os.path.exists(retool_file):
                retool = load_json(retool_file)
                self.retool = retool["variants"]

        # If we're using the RA hashes, pull it out here
        self.ra_hashes = ra_hashes
        if self.use_ra_hashes and self.ra_hashes is None:
            ra_hash_dir = self.config.get("dirs", {}).get("ra_hash_dir", None)
            if ra_hash_dir is None:
                raise ValueError("ra_hash_dir must be specified in config.yml")
            ra_hash_file = os.path.join(ra_hash_dir, f"{platform}.json")
            if os.path.exists(ra_hash_file):
                self.ra_hashes = load_json(ra_hash_file)

        self.hash_method = self.platform_config.get("ra_hash_method", None)

        self.log_line_sep = log_line_sep
        self.log_line_length = log_line_length

    def run(
        self,
        files,
    ):
        """Run the ROM parser"""

        game_dict = copy.deepcopy(files)

        self.logger.debug(f"{self.log_line_sep * self.log_line_length}")
        self.logger.debug(
            centred_string(
                f"Running ROMParser for {self.game}", total_length=self.log_line_length
            )
        )
        self.logger.debug(f"{self.log_line_sep * self.log_line_length}")

        for f in files:

            # # Get the potential title position out for compilations
            title_pos = files[f].get("title_pos", None)

            f_parsed = self.parse_file(
                f=f,
                file_dict=files[f],
                title_pos=title_pos,
            )
            game_dict[f].update(f_parsed)

        return game_dict

    def parse_file(
        self,
        f=None,
        file_dict=None,
        title_pos=None,
    ):
        """Parse useful info out of a specific file

        Args:
            f (str): Filename. Will only use this if something more suitable isn't found
            file_dict (dict): Dictionary of file properties
            title_pos (int, optional): Title position for compilations. Defaults to None.
        """

        if file_dict is None:
            file_dict = {}

        if self.use_filename:
            file_dict = self.parse_filename(
                f=f,
                file_dict=file_dict,
                title_pos=title_pos,
            )

        if self.use_retool:
            file_dict = self.parse_retool(file_dict=file_dict)

        if self.use_dat:
            file_dict = self.parse_dat(
                f=f,
                file_dict=file_dict,
            )

        file_dict["has_cheevos"] = False
        file_dict["patch_file"] = ""

        if self.use_ra_hashes:
            file_dict = self.parse_ra_hashes(
                f=f,
                file_dict=file_dict,
            )

        # Any last minute finalisations
        self.finalise_file_dict(file_dict)

        # File modification time
        full_file_path = os.path.join(
            self.raw_dir, self.platform, file_dict.get("original_name", f)
        )
        file_time = get_file_time(
            full_file_path,
            datetime_format=self.default_config["datetime_format"],
        )
        file_dict["file_mod_time"] = file_time

        # Log out these tags in a nice readable way
        self.logger.debug(centred_string(f"{f}:", total_length=self.log_line_length))

        # Track the various tags we can have
        true_tags = []
        false_tags = []
        none_tags = []
        str_tags = {}
        int_tags = {}
        list_tags = {}

        for key in file_dict:
            if isinstance(file_dict[key], bool):
                if file_dict[key]:
                    true_tags.append(key)
                else:
                    false_tags.append(key)
            elif isinstance(file_dict[key], str):
                str_tags[key] = file_dict[key]
            elif isinstance(file_dict[key], list):
                list_tags[key] = file_dict[key]
            elif isinstance(file_dict[key], int):
                int_tags[key] = file_dict[key]
            elif file_dict[key] is None:
                none_tags.append(key)
            else:
                raise ValueError(
                    f"{file_dict[key]} is not something I know how to parse"
                )

        # Log the string tags
        self.logger.debug(
            left_aligned_string(f"String tags:", total_length=self.log_line_length)
        )
        for tag in str_tags:
            if str_tags[tag] == "":
                continue
            self.logger.debug(
                left_aligned_string(
                    f"-> {tag}: {str_tags[tag]}", total_length=self.log_line_length
                )
            )

        # Log the list tags
        self.logger.debug(
            left_aligned_string(f"List tags:", total_length=self.log_line_length)
        )
        for tag in list_tags:
            if not list_tags[tag]:
                continue
            self.logger.debug(
                left_aligned_string(
                    f"-> {tag}: {', '.join(str(i) for i in list_tags[tag])}",
                    total_length=self.log_line_length,
                )
            )

        # Log the list tags
        self.logger.debug(
            left_aligned_string(f"Number tags:", total_length=self.log_line_length)
        )
        for tag in int_tags:
            self.logger.debug(
                left_aligned_string(
                    f"-> {tag}: {int_tags[tag]}",
                    total_length=self.log_line_length,
                )
            )

        # Log the True bool tags
        self.logger.debug(
            left_aligned_string(f"Tagged:", total_length=self.log_line_length)
        )
        for tag in true_tags:
            self.logger.debug(
                left_aligned_string(f"-> {tag}", total_length=self.log_line_length)
            )

        # Log the False bool tags
        self.logger.debug(
            left_aligned_string(f"Not tagged:", total_length=self.log_line_length)
        )
        for tag in false_tags:
            self.logger.debug(
                left_aligned_string(f"-> {tag}", total_length=self.log_line_length)
            )

        # Log any None tags
        self.logger.debug(
            left_aligned_string(f"None:", total_length=self.log_line_length)
        )
        for tag in none_tags:
            self.logger.debug(
                left_aligned_string(f"-> {tag}", total_length=self.log_line_length)
            )

        self.logger.debug(f"{'-' * self.log_line_length}")

        return file_dict

    def parse_retool(self, file_dict=None):
        """Parse info out of the retool file"""

        if file_dict is None:
            file_dict = {}

        if self.retool is None:
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            self.logger.warning(
                centred_string(
                    f"No retool file found for {self.platform}. Skipping",
                    total_length=self.log_line_length,
                )
            )
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            return file_dict

        # Loop over the variants, see if we get a match
        found_cat = False
        for retool_dict in self.retool:

            if found_cat:
                continue

            # If we don't have titles within the dupe dict, skip
            if "titles" not in retool_dict:
                continue

            # Match properly given search terms
            full_name = copy.deepcopy(file_dict["full_name"])
            short_name = copy.deepcopy(file_dict["short_name"])
            region_free_name = copy.deepcopy(file_dict["region_free_name"])

            for t in retool_dict["titles"]:
                search_term = t["searchTerm"]
                match_type = t.get("nameType", None)

                found_retool_variant = match_retool_search_terms(
                    full_name=full_name,
                    search_term=search_term,
                    short_name=short_name,
                    region_free_name=region_free_name,
                    match_type=match_type,
                )

                if found_retool_variant:
                    retool_cats = retool_dict.get("categories", [])
                    for retool_cat in retool_cats:
                        file_cat = retool_cat.lower().replace(" ", "_")
                        file_dict[file_cat] = True

        return file_dict

    def parse_dat(
        self,
        f=None,
        file_dict=None,
    ):
        """Parse info out of the dat file

        Args:
            f (str): Fallback filename
            file_dict (dict): Dictionary of file info
        """

        if file_dict is None:
            file_dict = {}

        f = copy.deepcopy(file_dict.get("original_name", f))

        if self.dat is None:
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            self.logger.warning(
                centred_string(
                    f"No dat file found for {self.platform}. Skipping",
                    total_length=self.log_line_length,
                )
            )
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            return file_dict

        # Remember there aren't zips in the dat entries
        dat_entry = self.dat.get(f.rstrip(".zip"), None)
        if not dat_entry:
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            self.logger.warning(
                centred_string(
                    f"No dat entry found for {f}. Skipping",
                    total_length=self.log_line_length,
                )
            )
            self.logger.warning(f"{self.log_line_sep * self.log_line_length}")
            return file_dict

        dat_categories = self.default_config.get("dat_categories", [])
        for dat_cat in dat_categories:

            dat_val = dat_entry.get("category", "")
            cat_val = dat_val == dat_cat

            dat_cat_dict = dat_cat.lower().replace(" ", "_")
            if dat_cat_dict in file_dict:
                file_dict[dat_cat_dict] = file_dict[dat_cat_dict] | cat_val
            else:
                file_dict[dat_cat_dict] = cat_val

        # Get the checksums out
        checksums = self.default_config.get("dat_checksums", [])
        for checksum in checksums:

            # Because sometimes we have multiple files within the ROM, loop over and append them all
            rom_entries = dat_entry.get("rom", [])
            if isinstance(rom_entries, dict):
                rom_entries = [rom_entries]

            for rom_entry in rom_entries:

                if checksum in rom_entry:
                    if checksum not in file_dict:
                        file_dict[checksum] = []

                    file_dict[checksum].append(rom_entry[checksum])

        return file_dict

    def parse_ra_hashes(
        self,
        f=None,
        file_dict=None,
    ):
        """See if we can find ROMs that support RetroAchievements"""

        # If we don't have a dictionary already, then we can't
        if file_dict is None:
            file_dict = {}

        f = copy.deepcopy(file_dict.get("original_name", f))

        if self.hash_method is None:
            self.logger.warning(
                centred_string(
                    f"RA hash method not defined for {self.platform}",
                    total_length=self.log_line_length,
                )
            )
            return file_dict

        file_dict = self.match_hashes(f, file_dict)

        return file_dict

    def match_hashes(
        self,
        f,
        file_dict,
    ):
        """Get whether ROM has cheevos by various potential hash methods

        Args:
            f (str): Filename
            file_dict (dict): Dictionary of ROM descriptions
        """

        has_cheevos = False
        patch_file = ""

        if self.hash_method == "md5":
            match_list = file_dict.get("md5", [])

        elif self.hash_method == "custom":
            match_list = [f.rstrip(".zip")]

        else:
            self.logger.warning(
                centred_string(
                    f"Cannot currently handle {self.hash_method} hash method",
                    total_length=self.log_line_length,
                )
            )
            return has_cheevos, patch_file

        # Get the potential RA match by name (this won't include potentially patched ROMs)
        has_cheevos, patch_file = self.get_ra_match(
            match_list=match_list,
        )

        # If we've found something, stop here
        if has_cheevos:
            file_dict["has_cheevos"] = has_cheevos
            file_dict["patch_file"] = patch_file
            return file_dict

        # If we're on a custom hash, and we haven't found anything, now look
        # via parsing the names
        if self.hash_method == "custom":
            has_cheevos, patch_file = self.get_parsed_match(
                match_list=match_list, want_patched_files=False
            )
            if has_cheevos:
                file_dict["has_cheevos"] = has_cheevos
                file_dict["patch_file"] = patch_file
                return file_dict

        # If we still haven't, now look through files to see if we just need
        # a patch (i.e. the hash will change, but we have the file)
        match_list = [f]
        has_cheevos, patch_file = self.get_parsed_match(
            match_list=match_list,
        )

        file_dict["has_cheevos"] = has_cheevos
        file_dict["patch_file"] = patch_file

        return file_dict

    def get_ra_match(
        self,
        match_list=None,
    ):
        """Match a file to RetroAchievements supported files

        Args:
            match_list (list): List of potential match patterns. Defaults to an empty list
        """

        has_cheevos = False
        patch_file = ""

        # Set up a list of potential matches
        if match_list is None:
            match_list = []
        # If we've got nothing, don't waste time
        if len(match_list) == 0:
            return has_cheevos, patch_file

        # Pull out the particular key we need
        if self.hash_method == "md5":
            key = "MD5"
        elif self.hash_method == "custom":
            key = "Name"
        else:
            self.logger.warning(
                centred_string(
                    f"Cannot currently handle {self.hash_method} hash method",
                    total_length=self.log_line_length,
                )
            )
            return has_cheevos, patch_file

        # Because of inconsistencies between naming schemes, just pull a huge dictionary out here rather than try
        # to be clever
        ra_dict = {}

        for r in self.ra_hashes:
            for h in self.ra_hashes[r]["Hashes"]:

                # If the RA list is a subset, then skip
                if is_ra_subset(r):
                    continue

                # Since these should be exact matches, skip those that have
                # patches
                if h["PatchUrl"] is not None:
                    continue

                # Also mark anything that's not pure No-Intro/Redump labelled
                if any([l not in self.ra_labels for l in h["Labels"]]):
                    continue

                # Use the md5 as the unique key, and then name as the thing we'll match to
                md5 = copy.deepcopy(h["MD5"])
                name = copy.deepcopy(h[key])
                name = name.strip()

                # If we're dealing with names, there might
                # be file extensions to strip
                if key in ["Name"]:
                    for ext in self.ra_file_exts:
                        if name.endswith(ext):
                            name = name.rstrip(ext)

                # FIXME: Here as a catch-all, hopefully won't be a problem
                if md5 in ra_dict:
                    raise ValueError(f"Hash {md5} multiply defined")

                ra_dict[md5] = {
                    "name": name,
                    "patch_file": h["PatchUrl"],
                }

        # Again, if there's nothing here just return
        if len(ra_dict) == 0:
            return has_cheevos, patch_file

        for m in match_list:
            for r in ra_dict:
                if m == ra_dict[r]["name"]:
                    has_cheevos = True
                    patch_file = ra_dict[r]["patch_file"]

        if patch_file is None:
            patch_file = ""

        return has_cheevos, patch_file

    def get_parsed_match(
        self,
        match_list=None,
        want_patched_files=True,
    ):
        """Match a file to RetroAchievements supported files that need patches

        Args:
            match_list (list): List of potential match patterns. Defaults to an empty list
            want_patched_files (bool): Whether we're looking for hashes with patches or not. Defaults to True
        """

        has_cheevos = False
        patch_file = ""

        # Set up a list of potential matches
        if match_list is None:
            match_list = []
        # If we've got nothing, don't waste time
        if len(match_list) == 0:
            return has_cheevos, patch_file

        # Because of inconsistencies between naming schemes, just pull a huge dictionary out here rather than try
        # to be clever
        ra_dict = {}
        for r in self.ra_hashes:
            for h in self.ra_hashes[r]["Hashes"]:

                # If the RA list is a subset, then skip
                if is_ra_subset(r):
                    continue

                # Are we looking for patch URLs or not?
                if want_patched_files:
                    if h["PatchUrl"] is None:
                        continue
                else:
                    if h["PatchUrl"] is not None:
                        continue

                # Use the md5 as the unique key, and then name as the thing we'll match to
                md5 = copy.deepcopy(h["MD5"])
                name = copy.deepcopy(h["Name"])
                name = name.strip()

                # There might be file extensions to strip
                for ext in self.ra_file_exts:
                    if name.endswith(ext):
                        name = name.rstrip(ext)

                # FIXME: Here as a catch-all, hopefully won't be a problem
                if md5 in ra_dict:
                    raise ValueError(f"Hash {md5} multiply defined")

                # Pull out the name before the first bracket to match later
                ra_dict[md5] = {
                    "name": name,
                    "short_name": name.split(" (")[0],
                    "patch_url": h["PatchUrl"],
                }

        # Again, if there's nothing here just return
        if len(ra_dict) == 0:
            return has_cheevos, patch_file

        multiple_patch_files_found = False

        for m in match_list:

            if multiple_patch_files_found:
                continue

            # Parse filename and get the short name
            m_parsed = self.parse_filename(f=m)
            m_short = m.split(" (")[0]

            for r in ra_dict:

                if multiple_patch_files_found:
                    continue

                # Match by short names to start
                if m_short == ra_dict[r]["short_name"]:

                    r_parsed = self.parse_filename(f=ra_dict[r]["name"])

                    # Force some version info in here, if the RA name doesn't have it
                    if r_parsed["version_no"] == "" and m_parsed["version_no"] != "":
                        if Version(m_parsed["version_no"]) == Version("1"):
                            r_parsed["version_no"] = copy.deepcopy(
                                m_parsed["version_no"]
                            )

                    # Now, make sure all the useful checks pass
                    ra_checks_passed = True
                    for check in self.ra_patch_checks:

                        # If we've already failed, then just skip
                        if not ra_checks_passed:
                            continue

                        ra_checks_passed = check_match(
                            m_parsed[check],
                            r_parsed[check],
                            checks_passed=ra_checks_passed,
                        )

                        # After this first pass, also see if any of the regex checks are grouped,
                        # and double-check the sublevel below. This is because we could have e.g.
                        # mismatched modern types (like a GameCube version vs a Wii U Virtual Console
                        # version), which inevitably won't match hashes
                        if ra_checks_passed:
                            if check not in self.regex_config:
                                for r_c in self.regex_config:

                                    if not ra_checks_passed:
                                        continue

                                    r_c_group = self.regex_config[r_c].get(
                                        "group", None
                                    )
                                    if r_c_group == check:
                                        ra_checks_passed = check_match(
                                            m_parsed[r_c],
                                            r_parsed[r_c],
                                            checks_passed=ra_checks_passed,
                                        )

                    if ra_checks_passed:

                        # If we seem to have multiple patch files defined,
                        # then raise a warning and assume there isn't a patch
                        if patch_file != "":
                            self.logger.warning(
                                centred_string(
                                    f"Multiple potential patch files found for {m}",
                                    total_length=self.log_line_length,
                                )
                            )
                            has_cheevos = False
                            patch_file = None
                            multiple_patch_files_found = True

                        else:
                            has_cheevos = True
                            patch_file = ra_dict[r]["patch_url"]

                        if patch_file is None:
                            patch_file = ""

        if patch_file is None:
            patch_file = ""

        return has_cheevos, patch_file

    def finalise_file_dict(
        self,
        file_dict,
    ):
        """Do any last minute finalisation to the file dict"""

        file_dict = self.set_game_category(file_dict)
        file_dict = self.set_implicit_languages(file_dict)

        return file_dict

    def set_game_category(
        self,
        file_dict,
    ):
        """If a dat category hasn't been set, set it to game"""

        dat_categories = self.default_config.get("dat_categories", [])

        for d in dat_categories:
            d_sanitized = d.lower().replace(" ", "_")

            if d_sanitized not in file_dict:
                file_dict[d_sanitized] = False

        if all(
            [file_dict[d.lower().replace(" ", "_")] is False for d in dat_categories]
        ):
            file_dict["games"] = True

        return file_dict

    def set_implicit_languages(
        self,
        file_dict,
    ):
        """Set implicit language from region, if we don't already have languages"""

        implied_languages = self.default_config.get("implied_languages", {})

        # Only set if languages is an empty list
        if not file_dict["languages"]:
            for r in file_dict["regions"]:
                if r in implied_languages:
                    file_dict["languages"].append(implied_languages[r])

        return file_dict

    def parse_filename(
        self,
        f=None,
        file_dict=None,
        title_pos=None,
    ):
        """Parse info out of filename

        Args:
            f (str): filename. Defaults to None, which will pull the original
                name out of the dict
            title_pos (int): Title position for compilations. Defaults to None
            file_dict (dict): Existing file dictionary. Defaults to None, which
                will create an empty one
        """

        if file_dict is None:
            file_dict = {}

        if "full_name" not in file_dict and f is None:
            raise ValueError(
                "Either f needs to be defined, or full_name needs to be in the file dictionary"
            )

        if f is None:
            # Pull the filename out, which is the full name
            f = copy.deepcopy(file_dict["full_name"])

        # Split file into tags
        tags = [f"({x}" for x in f.rstrip(".zip").split(" (")][1:]

        for regex_key in self.regex_config:

            # Are we potentially using the title position?
            use_title_pos = False
            if regex_key in USE_TITLE_POS:
                use_title_pos = True

            regex_type = self.regex_config[regex_key].get("type", "bool")
            search_tags = self.regex_config[regex_key].get("search_tags", True)
            group = self.regex_config[regex_key].get("group", None)
            regex_flags = self.regex_config[regex_key].get("flags", "I")
            transform_pattern = self.regex_config[regex_key].get(
                "transform_pattern", None
            )
            transform_repl = self.regex_config[regex_key].get("transform_repl", None)

            dict_default_val = DICT_DEFAULT_VALS.get(regex_type, None)
            if dict_default_val is None:
                raise ValueError(
                    f"regex_type should be one of {list(DICT_DEFAULT_VALS.keys())}"
                )

            if regex_key not in file_dict:
                file_dict[regex_key] = copy.deepcopy(dict_default_val)

            if regex_flags == "NOFLAG":
                regex_flags = re.NOFLAG
            elif regex_flags == "I":
                regex_flags = re.I
            else:
                raise ValueError("regex_flags should be one of 'NOFLAG', 'I'")

            pattern = self.regex_config[regex_key]["pattern"]

            pattern_mappings = None

            if regex_type == "list":

                if isinstance(self.default_config[regex_key], dict):
                    str_to_join = [
                        self.default_config[regex_key][key]
                        for key in self.default_config[regex_key]
                    ]
                    pattern_mappings = self.default_config[regex_key]
                else:
                    str_to_join = copy.deepcopy(self.default_config[regex_key])

                pattern = pattern.replace(f"[{regex_key}]", "|".join(str_to_join))

            regex = re.compile(pattern, flags=regex_flags)
            if search_tags:

                found_tag = False

                for tag in tags:

                    if found_tag:
                        continue

                    pattern_string = get_pattern_val(
                        regex,
                        tag,
                        regex_type,
                        pattern_mappings=pattern_mappings,
                        title_pos=title_pos,
                        use_title_pos=use_title_pos,
                    )
                    if pattern_string is not None:

                        if transform_pattern is not None:
                            pattern_string = re.sub(
                                transform_pattern, transform_repl, pattern_string
                            )

                        file_dict[regex_key] = pattern_string
                        found_tag = True
            else:
                pattern_string = get_pattern_val(
                    regex,
                    f,
                    regex_type,
                    pattern_mappings=pattern_mappings,
                    title_pos=title_pos,
                    use_title_pos=use_title_pos,
                )
                if pattern_string is not None:
                    file_dict[regex_key] = pattern_string

            # Update groups, if needed
            if group is not None:
                if group not in file_dict:
                    file_dict[group] = dict_default_val

                if regex_type == "bool":
                    file_dict[group] = file_dict[group] | file_dict[regex_key]
                elif regex_type == "str":
                    if file_dict[group] and file_dict[regex_key]:
                        raise ValueError("Can't combine multiple groups with type str")
                    else:
                        file_dict[group] += file_dict[regex_key]
                elif regex_type == "list":
                    file_dict[group].extend(file_dict[regex_key])
                else:
                    raise ValueError(
                        f"regex_type should be one of {list(DICT_DEFAULT_VALS.keys())}"
                    )

        return file_dict



from romsearch import ROMParser, ROMChooser

TEST_NAME = "Example Game (USA) (En,De,Fr)"


def test_romchooser_version():
    """Put a number of versions through and check ROMChooser gets the right one"""

    test_case = {
        TEST_NAME: {"priority": 1},
        f"{TEST_NAME} (v1.00)": {"priority": 1},
        f"{TEST_NAME} (v2.00)": {"priority": 1},
    }

    rp = ROMParser(
        config_file="test_config.yml",
        platform="Nintendo - Super Nintendo Entertainment System",
        game="Example Game",
    )
    rom_dict = rp.run(test_case)

    rc = ROMChooser(
        config_file="test_config.yml",
        platform="Nintendo - Super Nintendo Entertainment System",
        game="Example Game",
    )
    rom_dict = rc.run(rom_dict)

    roms_found = [r for r in rom_dict]

    assert roms_found == ["Example Game (USA) (En,De,Fr) (v2.00)"]



import copy
import os
import re
import time
from datetime import datetime


def split(full_list, chunk_size=10):
    """Split a list in chunks of size chunk_size

    Args:
        full_list (list): list to split
        chunk_size (int, optional): size of each chunk. Defaults to 10
    """

    for i in range(0, len(full_list), chunk_size):
        yield full_list[i : i + chunk_size]


def match_retool_search_terms(
    full_name,
    search_term,
    short_name=None,
    region_free_name=None,
    match_type=None,
):
    """Match a name against a search term, given retool's matching rules

    Args:
        full_name (str): Full name for the ROM
        search_term (str): Search term to match against
        short_name (str): Short name to match against. Defaults to None,
            which inherits the full name
        region_free_name (str): Region free name to match against. Defaults to None,
            which inherits the full name
        match_type (str): Type of matching. Defaults to None,
            which will match against short name
    """

    match_found = False

    # Assign default short/region-free names if not supplied
    if short_name is None:
        short_name = copy.deepcopy(full_name)
    if region_free_name is None:
        region_free_name = copy.deepcopy(full_name)

    # If none, match against lowercased short name
    if match_type is None:
        if short_name.lower() == search_term.lower():
            match_found = True

    # If full, match against lowercase full name
    elif match_type == "full":
        if full_name.lower() == search_term.lower():
            match_found = True

    # If region-free, match against lowercased region-free name
    elif match_type == "regionFree":
        if region_free_name.lower() == search_term.lower():
            match_found = True

    # If regex, match against full name
    elif match_type == "regex":
        match = re.search(search_term, full_name)

        if match is not None:
            match_found = True

    else:
        raise ValueError(f"Unsure how to deal with name type {match_type}")

    return match_found


def normalize_name(
    f,
    disc_rename=None,
):
    """Normalize a name to standard form

    Currently, just normalizes the disc name

    Args:
        f (str): Name to normalize
        disc_rename (dict, optional): Disc rename mappings. Defaults to None.
    """

    f_norm = copy.deepcopy(f)

    if disc_rename is not None:
        for k, v in disc_rename.items():
            if k in f_norm:
                f_norm = f_norm.replace(k, v)

    return f_norm


def get_file_time(
    f,
    datetime_format="%Y/%m/%d, %H:%M:%S",
    return_as_str=True,
):
    """Get created file time from the file itself

    Args:
        f (str): Filename
        datetime_format (str, optional): Date and time format. Defaults to "%Y/%m/%d %H:%M:%S"
        return_as_str (bool, optional): Return string or full datetime. Defaults to True
    """

    if os.path.exists(f):
        ti_m = os.path.getmtime(f)
        date_ti_m = datetime.strptime(time.ctime(ti_m), "%a %b %d %H:%M:%S %Y")
    else:
        date_ti_m = datetime(year=1900, month=1, day=1, hour=0, minute=0, second=0)

    if return_as_str:
        date_ti_m_str = date_ti_m.strftime(format=datetime_format)
        return date_ti_m_str
    else:
        return date_ti_m



from discordwebhook import Discord


def discord_push(
    url,
    name,
    fields,
):
    """Post a message to Discord"""

    discord = Discord(url=url)
    discord.post(
        embeds=[
            {
                "author": {
                    "name": name,
                    "url": "https://github.com/bbtufty/romsearch",
                },
                "fields": fields,
            }
        ],
    )

    return True



import copy
import glob
import os
import shutil

import romsearch
from .rompatcher import ROMPatcher
from ..util import (
    centred_string,
    load_yml,
    setup_logger,
    unzip_file,
    load_json,
    save_json,
)


class ROMMover:

    def __init__(
        self,
        platform,
        game,
        config_file=None,
        config=None,
        platform_config=None,
        logger=None,
        log_line_sep="=",
        log_line_length=100,
    ):
        """ROM Moving and cache updating tool

        Because we do this per-platform, per-game, they need to be specified here

        Args:
            platform (str): Platform name
            game (str): Game name
            config_file (str, optional): path to config file. Defaults to None.
            config (dict, optional): configuration dictionary. Defaults to None.
            platform_config (dict, optional): platform configuration dictionary. Defaults to None.
            logger (logging.Logger, optional): logger. Defaults to None.
            log_line_length (int, optional): Line length of log. Defaults to 100
        """

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get(
                "log_dir", os.path.join(os.getcwd(), "logs")
            )
            logger_add_dir = str(os.path.join(platform, game))
            log_level = self.config.get("logger", {}).get("level", "info")
            logger = setup_logger(
                log_level=log_level,
                script_name=f"ROMMover",
                log_dir=log_dir,
                additional_dir=logger_add_dir,
            )
        self.logger = logger

        # Pull in directories
        self.raw_dir = self.config.get("dirs", {}).get("raw_dir", None)
        if self.raw_dir is None:
            raise ValueError("raw_dir needs to be defined in config")

        self.rom_dir = self.config.get("dirs", {}).get("rom_dir", None)
        if self.rom_dir is None:
            raise ValueError("rom_dir needs to be defined in config")

        self.run_rompatcher = self.config.get("romsearch", {}).get(
            "run_rompatcher", False
        )
        self.patch_dir = self.config.get("dirs", {}).get("patch_dir", None)
        if self.patch_dir is None and self.run_rompatcher:
            raise ValueError("patch_dir needs to be defined in config")

        cache_dir = self.config.get("dirs", {}).get("cache_dir", os.getcwd())
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        cache_file = os.path.join(cache_dir, f"cache ({platform}).json")

        if os.path.exists(cache_file):
            cache = load_json(cache_file)
        else:
            cache = {}

        self.platform = platform
        self.game = game
        self.cache_file = cache_file
        self.cache = cache

        # Pull in platform config that we need
        mod_dir = os.path.dirname(romsearch.__file__)

        if platform_config is None:
            platform_config_file = os.path.join(
                mod_dir, "configs", "platforms", f"{platform}.yml"
            )
            platform_config = load_yml(platform_config_file)
        self.platform_config = platform_config

        self.unzip = self.platform_config.get("unzip", False)

        self.log_line_sep = log_line_sep
        self.log_line_length = log_line_length

    def run(
        self,
        rom_dict,
    ):

        roms_moved = self.move_roms(rom_dict)
        self.save_cache()

        return roms_moved

    def move_roms(self, rom_dict):
        """Actually move the roms"""

        roms_moved = []

        for rom_no, rom in enumerate(rom_dict):

            # Pull out a clean directory name
            dir_name = str(copy.deepcopy(rom_dict[rom]["dir_name"]))

            cache_mod_time = (
                self.cache.get(self.platform, {})
                .get(self.game, {})
                .get(rom, {})
                .get("file_mod_time", 0)
            )

            # Figure out if we're patching the ROM, based on whether it's already been patched
            # and if there's a patch file
            rom_patched = (
                self.cache.get(self.platform, {})
                .get(self.game, {})
                .get(rom, {})
                .get("patched", False)
            )
            patch_url = rom_dict.get(rom, {}).get("patch_file", "")

            to_patch_rom = False
            if not rom_patched and patch_url != "" and self.run_rompatcher:
                to_patch_rom = True

            # Keep track if we're expecting a patched file
            expecting_patched_rom = False
            if patch_url != "" and self.run_rompatcher:
                expecting_patched_rom = True

            # Skip if the file modification time matches the one in the cache, we're not patching, and the
            # destination file exists
            out_dir = os.path.join(self.rom_dir, self.platform, dir_name)

            # Loop over here, since the extensions might change
            out_files = glob.glob(os.path.join(out_dir, "*"))

            final_file_exists = False
            for o in out_files:

                if final_file_exists:
                    continue

                o_base = os.path.basename(o)
                o_short = os.path.splitext(o_base)[0]

                rom_short = os.path.splitext(rom)[0]
                if expecting_patched_rom:
                    rom_short += " (ROMPatched)"

                if o_short == rom_short:
                    final_file_exists = True

            if rom_dict[rom]["file_mod_time"] == cache_mod_time and final_file_exists:
                self.logger.info(
                    centred_string(
                        f"No updates for {rom}, skipping",
                        total_length=self.log_line_length,
                    )
                )
                continue

            # If we're patching ROMs, then do that here
            if to_patch_rom:

                rom_file = os.path.join(self.raw_dir, self.platform, rom)

                patcher = ROMPatcher(
                    platform=self.platform,
                    config=self.config,
                    logger=self.logger,
                    log_line_length=self.log_line_length,
                )

                full_rom = patcher.run(
                    file=rom_file,
                    patch_url=patch_url,
                )

                unzip = False
                patched = True

            else:
                full_dir = os.path.join(self.raw_dir, self.platform)
                full_rom = os.path.join(str(full_dir), rom)
                unzip = copy.deepcopy(self.unzip)

                patched = False

            # Log whether we've patched or not
            rom_dict[rom]["patched"] = patched

            # Move the main file. Don't delete folders as the game and the out directory
            # don't necessarily match
            move_file_success = self.move_file(full_rom, out_dir=out_dir, unzip=unzip)

            if not move_file_success:
                self.logger.warning(
                    centred_string(
                        f"{rom} not found in raw directory, skipping",
                        total_length=self.log_line_length,
                    )
                )
                continue

            self.logger.info(
                centred_string(f"Moved {rom}", total_length=self.log_line_length)
            )

            # If there are additional file to move/unzip, do that now
            if "additional_dirs" in self.platform_config:
                for add_dir in self.platform_config["additional_dirs"]:

                    add_full_dir = f"{self.raw_dir} {add_dir}"
                    add_file = os.path.join(add_full_dir, rom)
                    if os.path.exists(add_file):
                        self.move_file(add_file, unzip=self.unzip)
                        self.logger.info(
                            centred_string(
                                f"Moved {rom} {add_dir}",
                                total_length=self.log_line_length,
                            )
                        )

            # Update the cache
            self.cache_update(file=rom, rom_dict=rom_dict)

            roms_moved.append(rom)

        return roms_moved

    def move_file(
        self,
        zip_file_name,
        out_dir=None,
        unzip=False,
        delete_folder=False,
    ):
        """Move file to directory structure, optionally unzipping"""

        # If the file doesn't exist, crash out
        if not os.path.exists(zip_file_name):
            return False

        # If we don't have an output directory, set one from the game name
        if out_dir is None:
            out_dir = os.path.join(self.rom_dir, self.platform, self.game)

        if delete_folder and os.path.exists(out_dir):
            shutil.rmtree(out_dir)

        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        out_dir = str(out_dir)

        if unzip:
            unzip_file(zip_file_name, out_dir)
        else:
            short_zip_file = os.path.split(zip_file_name)[-1]
            out_file = os.path.join(out_dir, short_zip_file)

            # Remove this file if it already exists
            if os.path.exists(out_file):
                os.remove(out_file)

            os.link(zip_file_name, out_file)

        return True

    def cache_update(self, file, rom_dict):
        """Update the cache with new file data"""

        if self.platform not in self.cache:
            self.cache[self.platform] = {}

        if self.game not in self.cache[self.platform]:
            self.cache[self.platform][self.game] = {}

        # If there's already something in there, clear it out
        if not self.cache[self.platform][self.game]:
            self.cache[self.platform][self.game] = {}

        # Include info about whether the ROM has been patched or not,
        # and the patch file
        self.cache[self.platform][self.game][file] = {
            "file_mod_time": rom_dict[file]["file_mod_time"],
            "patch_file": rom_dict[file]["patch_file"],
            "patched": rom_dict[file]["patched"],
        }

    def save_cache(self):
        """Save out the cache file"""

        cache = copy.deepcopy(self.cache)
        save_json(cache, self.cache_file, sort_key=self.platform)
